# encoding=utf8

"""Particle swarm algorithm module."""

import numpy as np

from niapy.algorithms.algorithm import Algorithm
from niapy.util import full_array
from niapy.util.repair import reflect

__all__ = [
    'ParticleSwarmAlgorithm',
    'ParticleSwarmOptimization',
    'CenterParticleSwarmOptimization',
    'MutatedParticleSwarmOptimization',
    'MutatedCenterParticleSwarmOptimization',
    'ComprehensiveLearningParticleSwarmOptimizer',
    'MutatedCenterUnifiedParticleSwarmOptimization',
    'OppositionVelocityClampingParticleSwarmOptimization'
]


class ParticleSwarmAlgorithm(Algorithm):
    r"""Implementation of Particle Swarm Optimization algorithm.

    Algorithm:
        Particle Swarm Optimization algorithm

    Date:
        2018

    Authors:
        Lucija Brezočnik, Grega Vrbančič, Iztok Fister Jr. and Klemen Berkovič

    License:
        MIT

    Reference paper:
        Kennedy, J. and Eberhart, R. "Particle Swarm Optimization". Proceedings of IEEE International Conference on Neural Networks. IV. pp. 1942--1948, 1995.

    Attributes:
        Name (List[str]): List of strings representing algorithm names
        c1 (float): Cognitive component.
        c2 (float): Social component.
        w (Union[float, numpy.ndarray[float]]): Inertial weight.
        min_velocity (Union[float, numpy.ndarray[float]]): Minimal velocity.
        max_velocity (Union[float, numpy.ndarray[float]]): Maximal velocity.
        repair (Callable[[numpy.ndarray, numpy.ndarray, numpy.ndarray, Optional[numpy.random.Generator]], numpy.ndarray]): Repair method for velocity.

    See Also:
        * :class:`niapy.algorithms.Algorithm`

    """

    Name = ['WeightedVelocityClampingParticleSwarmAlgorithm', 'WVCPSO']

    @staticmethod
    def info():
        r"""Get basic information of algorithm.

        Returns:
            str: Basic information of algorithm.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""Kennedy, J. and Eberhart, R. "Particle Swarm Optimization". Proceedings of IEEE International Conference on Neural Networks. IV. pp. 1942--1948, 1995."""

    def __init__(self, population_size=25, c1=2.0, c2=2.0, w=0.7, min_velocity=-1.5, max_velocity=1.5, repair=reflect,
                 *args, **kwargs):
        """Initialize ParticleSwarmAlgorithm.

        Args:
            population_size (int): Population size
            c1 (float): Cognitive component.
            c2 (float): Social component.
            w (Union[float, numpy.ndarray]): Inertial weight.
            min_velocity (Union[float, numpy.ndarray]): Minimal velocity.
            max_velocity (Union[float, numpy.ndarray]): Maximal velocity.
            repair (Callable[[np.ndarray, np.ndarray, np.ndarray, dict], np.ndarray]): Repair method for velocity.

        See Also:
            * :func:`niapy.algorithms.Algorithm.__init__`

        """
        super().__init__(population_size, *args, **kwargs)
        self.c1 = c1
        self.c2 = c2
        self.w = w
        self.min_velocity = min_velocity
        self.max_velocity = max_velocity
        self.repair = repair

    def set_parameters(self, population_size=25, c1=2.0, c2=2.0, w=0.7, min_velocity=-1.5, max_velocity=1.5,
                       repair=reflect, **kwargs):
        r"""Set Particle Swarm Algorithm main parameters.

        Args:
            population_size (int): Population size
            c1 (float): Cognitive component.
            c2 (float): Social component.
            w (Union[float, numpy.ndarray]): Inertial weight.
            min_velocity (Union[float, numpy.ndarray]): Minimal velocity.
            max_velocity (Union[float, numpy.ndarray]): Maximal velocity.
            repair (Callable[[np.ndarray, np.ndarray, np.ndarray, dict], np.ndarray]): Repair method for velocity.

        See Also:
            * :func:`niapy.algorithms.Algorithm.set_parameters`

        """
        super().set_parameters(population_size=population_size, **kwargs)
        self.c1 = c1
        self.c2 = c2
        self.w = w
        self.min_velocity = min_velocity
        self.max_velocity = max_velocity
        self.repair = repair

    def get_parameters(self):
        r"""Get value of parameters for this instance of algorithm.

        Returns:
            Dict[str, Union[int, float, numpy.ndarray]]: Dictionary which has parameters mapped to values.

        See Also:
            * :func:`niapy.algorithms.Algorithm.get_parameters`

        """
        d = super().get_parameters()
        d.update({
            'c1': self.c1,
            'c2': self.c2,
            'w': self.w,
            'min_velocity': self.min_velocity,
            'max_velocity': self.max_velocity
        })
        return d

    def init(self, task):
        r"""Initialize dynamic arguments of Particle Swarm Optimization algorithm.

        Args:
            task (Task): Optimization task.

        Returns:
            Dict[str, Union[float, numpy.ndarray]]:
                * w (numpy.ndarray): Inertial weight.
                * min_velocity (numpy.ndarray): Minimal velocity.
                * max_velocity (numpy.ndarray): Maximal velocity.
                * v (numpy.ndarray): Initial velocity of particle.

        """
        return {
            'w': full_array(self.w, task.dimension),
            'min_velocity': full_array(self.min_velocity, task.dimension),
            'max_velocity': full_array(self.max_velocity, task.dimension),
            'v': np.zeros((self.population_size, task.dimension))
        }

    def init_population(self, task):
        r"""Initialize population and dynamic arguments of the Particle Swarm Optimization algorithm.

        Args:
            task: Optimization task.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, list, dict]:
                1. Initial population.
                2. Initial population fitness/function values.
                3. Additional arguments.
                4. Additional keyword arguments:
                    * personal_best (numpy.ndarray): particles best population.
                    * personal_best_fitness (numpy.ndarray[float]): particles best positions function/fitness value.
                    * w (numpy.ndarray): Inertial weight.
                    * min_velocity (numpy.ndarray): Minimal velocity.
                    * max_velocity (numpy.ndarray): Maximal velocity.
                    * v (numpy.ndarray): Initial velocity of particle.

        See Also:
            * :func:`niapy.algorithms.Algorithm.init_population`

        """
        pop, fpop, d = super().init_population(task)
        d.update(self.init(task))
        d.update({'personal_best': pop.copy(), 'personal_best_fitness': fpop.copy()})
        return pop, fpop, d

    def update_velocity(self, v, p, pb, gb, w, min_velocity, max_velocity, task, **kwargs):
        r"""Update particle velocity.

        Args:
            v (numpy.ndarray): Current velocity of particle.
            p (numpy.ndarray): Current position of particle.
            pb (numpy.ndarray): Personal best position of particle.
            gb (numpy.ndarray): Global best position of particle.
            w (Union[float, numpy.ndarray]): Weights for velocity adjustment.
            min_velocity (numpy.ndarray): Minimal velocity allowed.
            max_velocity (numpy.ndarray): Maximal velocity allowed.
            task (Task): Optimization task.
            kwargs: Additional arguments.

        Returns:
            numpy.ndarray: Updated velocity of particle.

        """
        return self.repair(
            w * v + self.c1 * self.random(task.dimension) * (pb - p) + self.c2 * self.random(task.dimension) * (gb - p),
            min_velocity, max_velocity)

    def run_iteration(self, task, pop, fpop, xb, fxb, **params):
        r"""Core function of Particle Swarm Optimization algorithm.

        Args:
            task (Task): Optimization task.
            pop (numpy.ndarray): Current populations.
            fpop (numpy.ndarray): Current population fitness/function values.
            xb (numpy.ndarray): Current best particle.
            fxb (float): Current best particle fitness/function value.
            params (dict): Additional function keyword arguments.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, float, dict]:
                1. New population.
                2. New population fitness/function values.
                3. New global best position.
                4. New global best positions function/fitness value.
                5. Additional arguments.
                6. Additional keyword arguments:
                    * personal_best (numpy.ndarray): Particles best population.
                    * personal_best_fitness (numpy.ndarray[float]): Particles best positions function/fitness value.
                    * w (numpy.ndarray): Inertial weight.
                    * min_velocity (numpy.ndarray): Minimal velocity.
                    * max_velocity (numpy.ndarray): Maximal velocity.
                    * v (numpy.ndarray): Initial velocity of particle.

        See Also:
            * :class:`niapy.algorithms.algorithm.Algorithm.run_iteration`

        """
        personal_best = params.pop('personal_best')
        personal_best_fitness = params.pop('personal_best_fitness')
        w = params.pop('w')
        min_velocity = params.pop('min_velocity')
        max_velocity = params.pop('max_velocity')
        v = params.pop('v')

        for i in range(len(pop)):
            v[i] = self.update_velocity(v[i], pop[i], personal_best[i], xb, w, min_velocity, max_velocity, task)
            pop[i] = task.repair(pop[i] + v[i], rng=self.rng)
            fpop[i] = task.eval(pop[i])
            if fpop[i] < personal_best_fitness[i]:
                personal_best[i], personal_best_fitness[i] = pop[i].copy(), fpop[i]
            if fpop[i] < fxb:
                xb, fxb = pop[i].copy(), fpop[i]
        return pop, fpop, xb, fxb, {'personal_best': personal_best, 'personal_best_fitness': personal_best_fitness,
                                    'w': w, 'min_velocity': min_velocity, 'max_velocity': max_velocity, 'v': v}


class ParticleSwarmOptimization(ParticleSwarmAlgorithm):
    r"""Implementation of Particle Swarm Optimization algorithm.

    Algorithm:
        Particle Swarm Optimization algorithm

    Date:
        2018

    Authors:
        Lucija Brezočnik, Grega Vrbančič, Iztok Fister Jr. and Klemen Berkovič

    License:
        MIT

    Reference paper:
        Kennedy, J. and Eberhart, R. "Particle Swarm Optimization". Proceedings of IEEE International Conference on Neural Networks. IV. pp. 1942--1948, 1995.

    Attributes:
        Name (List[str]): List of strings representing algorithm names

    See Also:
        * :class:`niapy.algorithms.basic.WeightedVelocityClampingParticleSwarmAlgorithm`

    """

    Name = ['ParticleSwarmAlgorithm', 'PSO']

    @staticmethod
    def info():
        r"""Get basic information of algorithm.

        Returns:
            str: Basic information of algorithm.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""Kennedy, J. and Eberhart, R. "Particle Swarm Optimization". Proceedings of IEEE International Conference on Neural Networks. IV. pp. 1942--1948, 1995."""

    def __init__(self, *args, **kwargs):
        """Initialize ParticleSwarmOptimization."""
        super().__init__(*args, **kwargs)
        self.w = 1.0
        self.min_velocity = -np.inf
        self.max_velocity = np.inf

    def set_parameters(self, **kwargs):
        r"""Set core parameters of algorithm.

        See Also:
            * :func:`niapy.algorithms.basic.WeightedVelocityClampingParticleSwarmAlgorithm.set_parameters`

        """
        kwargs.pop('w', None), kwargs.pop('vMin', None), kwargs.pop('vMax', None)
        super().set_parameters(w=1, min_velocity=-np.inf, max_velocity=np.inf, **kwargs)


class OppositionVelocityClampingParticleSwarmOptimization(ParticleSwarmAlgorithm):
    r"""Implementation of Opposition-Based Particle Swarm Optimization with Velocity Clamping.

    Algorithm:
        Opposition-Based Particle Swarm Optimization with Velocity Clamping

    Date:
        2019

    Authors:
        Klemen Berkovič

    License:
        MIT

    Reference paper:
        Shahzad, Farrukh, et al. "Opposition-based particle swarm optimization with velocity clamping (OVCPSO)." Advances in Computational Intelligence. Springer, Berlin, Heidelberg, 2009. 339-348

    Attributes:
        p0: Probability of opposite learning phase.
        w_min: Minimum inertial weight.
        w_max: Maximum inertial weight.
        sigma: Velocity scaling factor.

    See Also:
        * :class:`niapy.algorithms.basic.ParticleSwarmAlgorithm`

    """

    Name = ['OppositionVelocityClampingParticleSwarmOptimization', 'OVCPSO']

    @staticmethod
    def info():
        r"""Get basic information of algorithm.

        Returns:
            str: Basic information of algorithm.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""Shahzad, Farrukh, et al. "Opposition-based particle swarm optimization with velocity clamping (OVCPSO)." Advances in Computational Intelligence. Springer, Berlin, Heidelberg, 2009. 339-348"""

    def __init__(self, p0=.3, w_min=.4, w_max=.9, sigma=.1, c1=1.49612, c2=1.49612, *args, **kwargs):
        """Initialize OppositionVelocityClampingParticleSwarmOptimization.

        Args:
            p0 (float): Probability of running Opposite learning.
            w_min (numpy.ndarray): Minimal value of weights.
            w_max (numpy.ndarray): Maximum value of weights.
            sigma (numpy.ndarray): Velocity range factor.
            c1 (float): Cognitive component.
            c2 (float): Social component.

        See Also:
            * :func:`niapy.algorithm.basic.ParticleSwarmAlgorithm.__init__`

        """
        kwargs.pop('w', None)
        super().__init__(w=w_max, c1=c1, c2=c2, *args, **kwargs)
        self.p0 = p0
        self.w_min = w_min
        self.w_max = w_max
        self.sigma = sigma

    def set_parameters(self, p0=.3, w_min=.4, w_max=.9, sigma=.1, c1=1.49612, c2=1.49612, **kwargs):
        r"""Set core algorithm parameters.

        Args:
            p0 (float): Probability of running Opposite learning.
            w_min (numpy.ndarray): Minimal value of weights.
            w_max (numpy.ndarray): Maximum value of weights.
            sigma (numpy.ndarray): Velocity range factor.
            c1 (float): Cognitive component.
            c2 (float): Social component.

        See Also:
            * :func:`niapy.algorithm.basic.ParticleSwarmAlgorithm.set_parameters`

        """
        kwargs.pop('w', None)
        super().set_parameters(w=w_max, c1=c1, c2=c2, **kwargs)
        self.p0 = p0
        self.w_min = w_min
        self.w_max = w_max
        self.sigma = sigma

    def get_parameters(self):
        r"""Get value of parameters for this instance of algorithm.

        Returns:
            Dict[str, Union[int, float, numpy.ndarray]]: Dictionary which has parameters mapped to values.

        See Also:
            * :func:`niapy.algorithms.basic.ParticleSwarmAlgorithm.get_parameters`

        """
        d = ParticleSwarmAlgorithm.get_parameters(self)
        d.pop('min_velocity', None), d.pop('max_velocity', None)
        d.update({
            'p0': self.p0, 'w_min': self.w_min, 'w_max': self.w_max, 'sigma': self.sigma
        })
        return d

    @staticmethod
    def opposite_learning(s_l, s_h, pop, fpop, task):
        r"""Run opposite learning phase.

        Args:
            s_l (numpy.ndarray): lower limit of opposite particles.
            s_h (numpy.ndarray): upper limit of opposite particles.
            pop (numpy.ndarray): Current populations positions.
            fpop (numpy.ndarray): Current populations functions/fitness values.
            task (Task): Optimization task.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, float]:
                1. New particles position
                2. New particles function/fitness values
                3. New best position of opposite learning phase
                4. new best function/fitness value of opposite learning phase

        """
        s_r = s_l + s_h
        s = np.asarray([s_r - e for e in pop])
        s_f = np.asarray([task.eval(e) for e in s])
        s, s_f = np.concatenate([pop, s]), np.concatenate([fpop, s_f])
        sorted_indices = np.argsort(s_f)
        return s[sorted_indices[:len(pop)]], s_f[sorted_indices[:len(pop)]], s[sorted_indices[0]], s_f[
            sorted_indices[0]]

    def init_population(self, task):
        r"""Init starting population and dynamic parameters.

        Args:
            task (Task): Optimization task.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, list, dict]:
                1. Initialized population.
                2. Initialized populations function/fitness values.
                3. Additional arguments.
                4. Additional keyword arguments:
                    * personal_best (numpy.ndarray): particles best population.
                    * personal_best_fitness (numpy.ndarray[float]): particles best positions function/fitness value.
                    * vMin (numpy.ndarray): Minimal velocity.
                    * vMax (numpy.ndarray): Maximal velocity.
                    * V (numpy.ndarray): Initial velocity of particle.
                    * S_u (numpy.ndarray): upper bound for opposite learning.
                    * S_l (numpy.ndarray): lower bound for opposite learning.

        """
        pop, fpop, d = super().init_population(task)
        s_l, s_h = task.lower, task.upper
        pop, fpop, _, _ = self.opposite_learning(s_l, s_h, pop, fpop, task)
        pb_indices = np.where(fpop < d['personal_best_fitness'])
        d['personal_best'][pb_indices], d['personal_best_fitness'][pb_indices] = pop[pb_indices], fpop[pb_indices]
        d['min_velocity'], d['max_velocity'] = self.sigma * (task.upper - task.lower), self.sigma * (
                task.lower - task.upper)
        d.update({'s_l': s_l, 's_h': s_h})
        return pop, fpop, d

    def run_iteration(self, task, pop, fpop, xb, fxb, **params):
        r"""Core function of Opposite-based Particle Swarm Optimization with velocity clamping algorithm.

        Args:
            task (Task): Optimization task.
            pop (numpy.ndarray): Current population.
            fpop (numpy.ndarray): Current populations function/fitness values.
            xb (numpy.ndarray): Current global best position.
            fxb (float): Current global best positions function/fitness value.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, float, list, dict]:
                1. New population.
                2. New populations function/fitness values.
                3. New global best position.
                4. New global best positions function/fitness value.
                5. Additional arguments.
                6. Additional keyword arguments:
                    * personal_best: particles best population.
                    * personal_best_fitness: particles best positions function/fitness value.
                    * min_velocity: Minimal velocity.
                    * max_velocity: Maximal velocity.
                    * v: Initial velocity of particle.
                    * s_h: upper bound for opposite learning.
                    * s_l: lower bound for opposite learning.

        """
        personal_best = params.pop('personal_best')
        personal_best_fitness = params.pop('personal_best_fitness')
        min_velocity = params.pop('min_velocity')
        max_velocity = params.pop('max_velocity')
        v = params.pop('v')
        s_l = params.pop('s_l')
        s_h = params.pop('s_h')

        if self.random() < self.p0:
            pop, fpop, nb, fnb = self.opposite_learning(s_l, s_h, pop, fpop, task)
            pb_indices = np.where(fpop < personal_best_fitness)
            personal_best[pb_indices], personal_best_fitness[pb_indices] = pop[pb_indices], fpop[pb_indices]
            if fnb < fxb:
                xb, fxb = nb.copy(), fnb
        else:
            w = self.w_max - ((self.w_max - self.w_min) / task.max_iters) * (task.iters + 1)
            for i in range(len(pop)):
                v[i] = self.update_velocity(v[i], pop[i], personal_best[i], xb, w, min_velocity, max_velocity, task)
                pop[i] = task.repair(pop[i] + v[i], rng=self.rng)
                fpop[i] = task.eval(pop[i])
                if fpop[i] < personal_best_fitness[i]:
                    personal_best[i], personal_best_fitness[i] = pop[i].copy(), fpop[i]
                    if fpop[i] < fxb:
                        xb, fxb = pop[i].copy(), fpop[i]
            min_velocity, max_velocity = self.sigma * np.min(pop, axis=0), self.sigma * np.max(pop, axis=0)
        return pop, fpop, xb, fxb, {'personal_best': personal_best, 'personal_best_fitness': personal_best_fitness,
                                    'min_velocity': min_velocity,
                                    'max_velocity': max_velocity, 'v': v, 's_l': s_l, 's_h': s_h}


class CenterParticleSwarmOptimization(ParticleSwarmAlgorithm):
    r"""Implementation of Center Particle Swarm Optimization.

    Algorithm:
        Center Particle Swarm Optimization

    Date:
        2019

    Authors:
        Klemen Berkovič

    License:
        MIT

    Reference paper:
        H.-C. Tsai, Predicting strengths of concrete-type specimens using hybrid multilayer perceptrons with center-Unified particle swarm optimization, Adv. Eng. Softw. 37 (2010) 1104–1112.

    See Also:
        * :class:`niapy.algorithms.basic.WeightedVelocityClampingParticleSwarmAlgorithm`

    """

    Name = ['CenterParticleSwarmOptimization', 'CPSO']

    @staticmethod
    def info():
        r"""Get basic information of algorithm.

        Returns:
            str: Basic information of algorithm.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""H.-C. Tsai, Predicting strengths of concrete-type specimens using hybrid multilayer perceptrons with center-Unified particle swarm optimization, Adv. Eng. Softw. 37 (2010) 1104–1112."""

    def __init__(self, *args, **kwargs):
        """Initialize CPSO."""
        kwargs.pop('min_velocity', None), kwargs.pop('max_velocity', None)
        super().__init__(min_velocity=-np.inf, max_velocity=np.inf, *args, **kwargs)

    def set_parameters(self, **kwargs):
        r"""Set core algorithm parameters.

        Args:
            **kwargs: Additional arguments.

        See Also:
            :func:`niapy.algorithm.basic.WeightedVelocityClampingParticleSwarmAlgorithm.set_parameters`

        """
        kwargs.pop('min_velocity', None), kwargs.pop('max_velocity', None)
        super().set_parameters(min_velocity=-np.inf, max_velocity=np.inf, **kwargs)

    def get_parameters(self):
        r"""Get value of parameters for this instance of algorithm.

        Returns:
            Dict[str, Union[int, float, numpy.ndarray]]: Dictionary which has parameters mapped to values.

        See Also:
            * :func:`niapy.algorithms.basic.ParticleSwarmAlgorithm.get_parameters`

        """
        d = super().get_parameters()
        d.pop('min_velocity', None), d.pop('max_velocity', None)
        return d

    def run_iteration(self, task, pop, fpop, xb, fxb, **params):
        r"""Core function of algorithm.

        Args:
            task (Task): Optimization task.
            pop (numpy.ndarray): Current population of particles.
            fpop (numpy.ndarray): Current particles function/fitness values.
            xb (numpy.ndarray): Current global best particle.
            fxb (numpy.float): Current global best particles function/fitness value.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, float, dict]:
                1. New population of particles.
                2. New populations function/fitness values.
                3. New global best particle.
                4. New global best particle function/fitness value.
                5. Additional arguments.
                6. Additional keyword arguments.

        See Also:
            * :func:`niapy.algorithm.basic.WeightedVelocityClampingParticleSwarmAlgorithm.run_iteration`

        """
        pop, fpop, xb, fxb, d = super().run_iteration(task, pop, fpop, xb, fxb, **params)
        c = np.sum(pop, axis=0) / len(pop)
        fc = task.eval(c)
        if fc <= fxb:
            xb, fxb = c, fc
        return pop, fpop, xb, fxb, d


class MutatedParticleSwarmOptimization(ParticleSwarmAlgorithm):
    r"""Implementation of Mutated Particle Swarm Optimization.

    Algorithm:
        Mutated Particle Swarm Optimization

    Date:
        2019

    Authors:
        Klemen Berkovič

    License:
        MIT

    Reference paper:
        H. Wang, C. Li, Y. Liu, S. Zeng, a hybrid particle swarm algorithm with cauchy mutation, Proceedings of the 2007 IEEE Swarm Intelligence Symposium (2007) 356–360.

    Attributes:
        num_mutations (int): Number of mutations of global best particle.

    See Also:
        * :class:`niapy.algorithms.basic.WeightedVelocityClampingParticleSwarmAlgorithm`

    """

    Name = ['MutatedParticleSwarmOptimization', 'MPSO']

    @staticmethod
    def info():
        r"""Get basic information of algorithm.

        Returns:
            str: Basic information of algorithm.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""H. Wang, C. Li, Y. Liu, S. Zeng, a hybrid particle swarm algorithm with cauchy mutation, Proceedings of the 2007 IEEE Swarm Intelligence Symposium (2007) 356–360."""

    def __init__(self, num_mutations=10, *args, **kwargs):
        """Initialize MPSO."""
        kwargs.pop('min_velocity', None), kwargs.pop('max_velocity', None)
        super().__init__(min_velocity=-np.inf, max_velocity=np.inf, *args, **kwargs)
        self.num_mutations = num_mutations

    def set_parameters(self, num_mutations=10, **kwargs):
        r"""Set core algorithm parameters.

        Args:
            num_mutations (int): Number of mutations of global best particle.
            **kwargs: Additional arguments.

        See Also:
            * :func:`niapy.algorithm.basic.WeightedVelocityClampingParticleSwarmAlgorithm.set_parameters`

        """
        kwargs.pop('min_velocity', None), kwargs.pop('max_velocity', None)
        ParticleSwarmAlgorithm.set_parameters(self, min_velocity=-np.inf, max_velocity=np.inf, **kwargs)
        self.num_mutations = num_mutations

    def get_parameters(self):
        r"""Get value of parameters for this instance of algorithm.

        Returns:
            Dict[str, Union[int, float, numpy.ndarray]]: Dictionary which has parameters mapped to values.

        See Also:
            * :func:`niapy.algorithms.basic.ParticleSwarmAlgorithm.get_parameters`

        """
        d = ParticleSwarmAlgorithm.get_parameters(self)
        d.pop('min_velocity', None), d.pop('max_velocity', None)
        d.update({'num_mutations': self.num_mutations})
        return d

    def run_iteration(self, task, pop, fpop, xb, fxb, **params):
        r"""Core function of algorithm.

        Args:
            task (Task): Optimization task.
            pop (numpy.ndarray): Current population of particles.
            fpop (numpy.ndarray): Current particles function/fitness values.
            xb (numpy.ndarray): Current global best particle.
            fxb (float): Current global best particles function/fitness value.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, float, list, dict]:
                1. New population of particles.
                2. New populations function/fitness values.
                3. New global best particle.
                4. New global best particle function/fitness value.
                5. Additional arguments.
                6. Additional keyword arguments.

        See Also:
            * :func:`niapy.algorithm.basic.WeightedVelocityClampingParticleSwarmAlgorithm.run_iteration`

        """
        pop, fpop, xb, fxb, d = ParticleSwarmAlgorithm.run_iteration(self, task, pop, fpop, xb, fxb, **params)
        v = d['v']
        v_a = (np.sum(v, axis=0) / len(v))
        v_a = v_a / np.max(np.abs(v_a))
        for _ in range(self.num_mutations):
            g = task.repair(xb + v_a * self.uniform(task.lower, task.upper), self.rng)
            fg = task.eval(g)
            if fg <= fxb:
                xb, fxb = g, fg
        return pop, fpop, xb, fxb, d


class MutatedCenterParticleSwarmOptimization(CenterParticleSwarmOptimization):
    r"""Implementation of Mutated Particle Swarm Optimization.

    Algorithm:
        Mutated Center Particle Swarm Optimization

    Date:
        2019

    Authors:
        Klemen Berkovič

    License:
        MIT

    Reference paper:
        TODO find one

    Attributes:
        num_mutations (int): Number of mutations of global best particle.

    See Also:
        * :class:`niapy.algorithms.basic.CenterParticleSwarmOptimization`

    """

    Name = ['MutatedCenterParticleSwarmOptimization', 'MCPSO']

    @staticmethod
    def info():
        r"""Get basic information of algorithm.

        Returns:
            str: Basic information of algorithm.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""TODO find one"""

    def __init__(self, num_mutations=10, *args, **kwargs):
        """Initialize MCPSO."""
        kwargs.pop('min_velocity', None), kwargs.pop('max_velocity', None)
        super().__init__(min_velocity=-np.inf, max_velocity=np.inf, *args, **kwargs)
        self.num_mutations = num_mutations

    def set_parameters(self, num_mutations=10, **kwargs):
        r"""Set core algorithm parameters.

        Args:
            num_mutations (int): Number of mutations of global best particle.
            **kwargs: Additional arguments.

        See Also:
            * :func:`niapy.algorithm.basic.CenterParticleSwarmOptimization.set_parameters`

        """
        kwargs.pop('min_velocity', None), kwargs.pop('max_velocity', None)
        ParticleSwarmAlgorithm.set_parameters(self, min_velocity=-np.inf, max_velocity=np.inf, **kwargs)
        self.num_mutations = num_mutations

    def get_parameters(self):
        r"""Get value of parameters for this instance of algorithm.

        Returns:
            Dict[str, Union[int, float, numpy.ndarray]]: Dictionary which has parameters mapped to values.

        See Also:
            * :func:`niapy.algorithms.basic.CenterParticleSwarmOptimization.get_parameters`

        """
        d = CenterParticleSwarmOptimization.get_parameters(self)
        d.update({'num_mutations': self.num_mutations})
        return d

    def run_iteration(self, task, pop, fpop, xb, fxb, **params):
        r"""Core function of algorithm.

        Args:
            task (Task): Optimization task.
            pop (numpy.ndarray): Current population of particles.
            fpop (numpy.ndarray): Current particles function/fitness values.
            xb (numpy.ndarray): Current global best particle.
            fxb (float: Current global best particles function/fitness value.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, float, list, dict]:
                1. New population of particles.
                2. New populations function/fitness values.
                3. New global best particle.
                4. New global best particle function/fitness value.
                5. Additional arguments.
                6. Additional keyword arguments.

        See Also:
            * :func:`niapy.algorithm.basic.WeightedVelocityClampingParticleSwarmAlgorithm.run_iteration`

        """
        pop, fpop, xb, fxb, d = CenterParticleSwarmOptimization.run_iteration(self, task, pop, fpop, xb, fxb, **params)
        v = d['v']
        v_a = (np.sum(v, axis=0) / len(v))
        v_a = v_a / np.max(np.abs(v_a))
        for _ in range(self.num_mutations):
            g = task.repair(xb + v_a * self.uniform(task.lower, task.upper), self.rng)
            fg = task.eval(g)
            if fg <= fxb:
                xb, fxb = g, fg
        return pop, fpop, xb, fxb, d


class MutatedCenterUnifiedParticleSwarmOptimization(MutatedCenterParticleSwarmOptimization):
    r"""Implementation of Mutated Particle Swarm Optimization.

    Algorithm:
        Mutated Center Unified Particle Swarm Optimization

    Date:
        2019

    Authors:
        Klemen Berkovič

    License:
        MIT

    Reference paper:
        Tsai, Hsing-Chih. "Unified particle swarm delivers high efficiency to particle swarm optimization." Applied Soft Computing 55 (2017): 371-383.

    Attributes:
        Name (List[str]): Names of algorithm.

    See Also:
        * :class:`niapy.algorithms.basic.CenterParticleSwarmOptimization`

    """

    Name = ['MutatedCenterUnifiedParticleSwarmOptimization', 'MCUPSO']

    @staticmethod
    def info():
        r"""Get basic information of algorithm.

        Returns:
            str: Basic information of algorithm.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""Tsai, Hsing-Chih. "Unified particle swarm delivers high efficiency to particle swarm optimization." Applied Soft Computing 55 (2017): 371-383."""

    def update_velocity(self, v, p, pb, gb, w, min_velocity, max_velocity, task, **kwargs):
        r"""Update particle velocity.

        Args:
            v (numpy.ndarray): Current velocity of particle.
            p (numpy.ndarray): Current position of particle.
            pb (numpy.ndarray): Personal best position of particle.
            gb (numpy.ndarray): Global best position of particle.
            w (numpy.ndarray): Weights for velocity adjustment.
            min_velocity (numpy.ndarray): Minimal velocity allowed.
            max_velocity (numpy.ndarray): Maximal velocity allowed.
            task (Task): Optimization task.
            kwargs (dict): Additional arguments.

        Returns:
            numpy.ndarray: Updated velocity of particle.

        """
        r3 = self.random(task.dimension)
        return self.repair(
            w * v + self.c1 * self.random(task.dimension) * (pb - p) * r3 + self.c2 * self.random(task.dimension) * (
                        gb - p) * (1 - r3),
            min_velocity, max_velocity)


class ComprehensiveLearningParticleSwarmOptimizer(ParticleSwarmAlgorithm):
    r"""Implementation of Mutated Particle Swarm Optimization.

    Algorithm:
        Comprehensive Learning Particle Swarm Optimizer

    Date:
        2019

    Authors:
        Klemen Berkovič

    License:
        MIT

    Reference paper:
        J. J. Liang, a. K. Qin, P. N. Suganthan and S. Baskar, "Comprehensive learning particle swarm optimizer for global optimization of multimodal functions," in IEEE Transactions on Evolutionary Computation, vol. 10, no. 3, pp. 281-295, June 2006. doi: 10.1109/TEVC.2005.857610

    Reference URL:
        http://ieeexplore.ieee.org/stamp/stamp.jsp?tp=&arnumber=1637688&isnumber=34326

    Attributes:
        w0 (float): Inertia weight.
        w1 (float): Inertia weight.
        c (float): Velocity constant.
        m (int): Refresh rate.

    See Also:
        * :class:`niapy.algorithms.basic.ParticleSwarmAlgorithm`

    """

    Name = ['ComprehensiveLearningParticleSwarmOptimizer', 'CLPSO']

    @staticmethod
    def info():
        r"""Get basic information of algorithm.

        Returns:
            str: Basic information of algorithm.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""J. J. Liang, a. K. Qin, P. N. Suganthan and S. Baskar, "Comprehensive learning particle swarm optimizer for global optimization of multimodal functions," in IEEE Transactions on Evolutionary Computation, vol. 10, no. 3, pp. 281-295, June 2006. doi: 10.1109/TEVC.2005.857610	"""

    def __init__(self, m=10, w0=.9, w1=.4, c=1.49445, *args, **kwargs):
        """Initialize CLPSO."""
        super().__init__(*args, **kwargs)
        self.m = m
        self.w0 = w0
        self.w1 = w1
        self.c = c

    def set_parameters(self, m=10, w0=.9, w1=.4, c=1.49445, **kwargs):
        r"""Set Particle Swarm Algorithm main parameters.

        Args:
            w0 (int): Inertia weight.
            w1 (float): Inertia weight.
            c (float): Velocity constant.
            m (float): Refresh rate.
            kwargs (dict): Additional arguments

        See Also:
            * :func:`niapy.algorithms.basic.ParticleSwarmAlgorithm.set_parameters`

        """
        ParticleSwarmAlgorithm.set_parameters(self, **kwargs)
        self.m = m
        self.w0 = w0
        self.w1 = w1
        self.c = c

    def get_parameters(self):
        r"""Get value of parameters for this instance of algorithm.

        Returns:
            Dict[str, Union[int, float, numpy.ndarray]]: Dictionary which has parameters mapped to values.

        See Also:
            * :func:`niapy.algorithms.basic.ParticleSwarmAlgorithm.get_parameters`

        """
        d = ParticleSwarmAlgorithm.get_parameters(self)
        d.update({
            'm': self.m,
            'w0': self.w0,
            'w1': self.w1,
            'c': self.c
        })
        return d

    def init(self, task):
        r"""Initialize dynamic arguments of Particle Swarm Optimization algorithm.

        Args:
            task (Task): Optimization task.

        Returns:
            Dict[str, numpy.ndarray]:
                * vMin: Minimal velocity.
                * vMax: Maximal velocity.
                * V: Initial velocity of particle.
                * flag: Refresh gap counter.

        """
        return {'min_velocity': full_array(self.min_velocity, task.dimension),
                'max_velocity': full_array(self.max_velocity, task.dimension),
                'v': np.full([self.population_size, task.dimension], 0.0), 'flag': np.full(self.population_size, 0),
                'pc': np.asarray(
                    [.05 + .45 * (np.exp(10 * (i - 1) / (self.population_size - 1)) - 1) / (np.exp(10) - 1) for i in
                     range(self.population_size)])}

    def generate_personal_best_cl(self, i, pc, personal_best, personal_best_fitness):
        r"""Generate new personal best position for learning.

        Args:
            i (int): Current particle.
            pc (float): Learning probability.
            personal_best (numpy.ndarray): Personal best positions for population.
            personal_best_fitness (numpy.ndarray): Personal best positions function/fitness values for personal best position.

        Returns:
            numpy.ndarray: Personal best for learning.

        """
        pbest = []
        for j in range(len(personal_best[i])):
            if self.random() > pc:
                pbest.append(personal_best[i, j])
            else:
                r1, r2 = int(self.random() * len(personal_best)), int(self.random() * len(personal_best))
                if personal_best_fitness[r1] < personal_best_fitness[r2]:
                    pbest.append(personal_best[r1, j])
                else:
                    pbest.append(personal_best[r2, j])
        return np.asarray(pbest)

    def update_velocity_cl(self, v, p, pb, w, min_velocity, max_velocity, task, **_kwargs):
        r"""Update particle velocity.

        Args:
            v (numpy.ndarray): Current velocity of particle.
            p (numpy.ndarray): Current position of particle.
            pb (numpy.ndarray): Personal best position of particle.
            w (numpy.ndarray): Weights for velocity adjustment.
            min_velocity (numpy.ndarray): Minimal velocity allowed.
            max_velocity (numpy.ndarray): Maximal velocity allowed.
            task (Task): Optimization task.

        Returns:
            numpy.ndarray: Updated velocity of particle.

        """
        return self.repair(w * v + self.c * self.random(task.dimension) * (pb - p), min_velocity, max_velocity)

    def run_iteration(self, task, pop, fpop, xb, fxb, **params):
        r"""Core function of algorithm.

        Args:
            task (Task): Optimization task.
            pop (numpy.ndarray): Current populations.
            fpop (numpy.ndarray): Current population fitness/function values.
            xb (numpy.ndarray): Current best particle.
            fxb (float): Current best particle fitness/function value.
            params (dict): Additional function keyword arguments.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, list, dict]:
                1. New population.
                2. New population fitness/function values.
                3. New global best position.
                4. New global best positions function/fitness value.
                5. Additional arguments.
                6. Additional keyword arguments:
                    * personal_best: Particles best population.
                    * personal_best_fitness: Particles best positions function/fitness value.
                    * min_velocity: Minimal velocity.
                    * max_velocity: Maximal velocity.
                    * V: Initial velocity of particle.
                    * flag: Refresh gap counter.
                    * pc: Learning rate.

        See Also:
            * :class:`niapy.algorithms.basic.ParticleSwarmAlgorithm.run_iteration`

        """
        personal_best = params.pop('personal_best')
        personal_best_fitness = params.pop('personal_best_fitness')
        min_velocity = params.pop('min_velocity')
        max_velocity = params.pop('max_velocity')
        v = params.pop('v')
        flag = params.pop('flag')
        pc = params.pop('pc')

        w = self.w0 * (self.w0 - self.w1) * (task.iters + 1) / task.max_iters
        for i in range(len(pop)):
            if flag[i] >= self.m:
                v[i] = self.update_velocity(v[i], pop[i], personal_best[i], xb, 1, min_velocity, max_velocity, task)
                pop[i] = task.repair(pop[i] + v[i], rng=self.rng)
                fpop[i] = task.eval(pop[i])
                if fpop[i] < personal_best_fitness[i]:
                    personal_best[i], personal_best_fitness[i] = pop[i].copy(), fpop[i]
                    if fpop[i] < fxb:
                        xb, fxb = pop[i].copy(), fpop[i]
                flag[i] = 0
            pbest = self.generate_personal_best_cl(i, pc[i], personal_best, personal_best_fitness)
            v[i] = self.update_velocity_cl(v[i], pop[i], pbest, w, min_velocity, max_velocity, task)
            pop[i] = pop[i] + v[i]
            if task.is_feasible(pop[i]):
                fpop[i] = task.eval(pop[i])
                if fpop[i] < personal_best_fitness[i]:
                    personal_best[i], personal_best_fitness[i] = pop[i].copy(), fpop[i]
                    if fpop[i] < fxb:
                        xb, fxb = pop[i].copy(), fpop[i]
        return pop, fpop, xb, fxb, {'personal_best': personal_best, 'personal_best_fitness': personal_best_fitness,
                                    'min_velocity': min_velocity,
                                    'max_velocity': max_velocity, 'v': v, 'flag': flag, 'pc': pc}



# encoding=utf8
import logging
import math

import numpy as np

from niapy.algorithms.algorithm import Algorithm, Individual, default_individual_init
from niapy.util.array import objects_to_array

__all__ = ['DifferentialEvolution', 'DynNpDifferentialEvolution', 'AgingNpDifferentialEvolution',
           'MultiStrategyDifferentialEvolution', 'DynNpMultiStrategyDifferentialEvolution', 'AgingIndividual',
           'cross_rand1', 'cross_rand2', 'cross_best2', 'cross_best1', 'cross_best2', 'cross_curr2rand1',
           'cross_curr2best1', 'multi_mutations', 'proportional', 'linear', 'bilinear']

logging.basicConfig()
logger = logging.getLogger('niapy.algorithms.basic')
logger.setLevel('INFO')


def cross_rand1(pop, ic, f, cr, rng, **_kwargs):
    r"""Mutation strategy with crossover.

    Mutation strategy uses three different random individuals from population to perform mutation.

    Mutation:
        Name: DE/rand/1

        :math:`\mathbf{x}_{r_1, G} + differential_weight \cdot (\mathbf{x}_{r_2, G} - \mathbf{x}_{r_3, G}`
        where :math:`r_1, r_2, r_3` are random indexes representing current population individuals.

    Crossover:
        Name: Binomial crossover

        :math:`\mathbf{x}_{i, G+1} = \begin{cases} \mathbf{u}_{i, G+1}, & \text{if $f(\mathbf{u}_{i, G+1}) \leq f(\mathbf{x}_{i, G})$}, \\ \mathbf{x}_{i, G}, & \text{otherwise}. \end{cases}`

    Args:
        pop (numpy.ndarray[Individual]): Current population.
        ic (int): Index of individual being mutated.
        f (float): Scale factor.
        cr (float): Crossover probability.
        rng (numpy.random.Generator): Random generator.

    Returns:
        numpy.ndarray: Mutated and mixed individual.

    """
    j = rng.integers(len(pop[ic]))
    p = [1 / (len(pop) - 1.0) if i != ic else 0 for i in range(len(pop))] if len(pop) > 3 else None
    r = rng.choice(len(pop), 3, replace=not len(pop) >= 3, p=p)
    x = [pop[r[0]][i] + f * (pop[r[1]][i] - pop[r[2]][i]) if rng.random() < cr or i == j else pop[ic][i] for i in
         range(len(pop[ic]))]
    return np.asarray(x)


def cross_best1(pop, ic, f, cr, rng, x_b=None, **_kwargs):
    r"""Mutation strategy with crossover.

    Mutation strategy uses two different random individuals from population and global best individual.

    Mutation:
        Name: de/best/1

        :math:`\mathbf{v}_{i, G} = \mathbf{x}_{best, G} + differential_weight \cdot (\mathbf{x}_{r_1, G} - \mathbf{x}_{r_2, G})`
        where :math:`r_1, r_2` are random indexes representing current population individuals.

    Crossover:
        Name: Binomial crossover

        :math:`\mathbf{x}_{i, G+1} = \begin{cases} \mathbf{u}_{i, G+1}, & \text{if $f(\mathbf{u}_{i, G+1}) \leq f(\mathbf{x}_{i, G})$}, \\ \mathbf{x}_{i, G}, & \text{otherwise}. \end{cases}`

    args:
        pop (numpy.ndarray[Individual]): Current population.
        ic (int): Index of individual being mutated.
        x_b (Individual): Current global best individual.
        f (float): Scale factor.
        cr (float): Crossover probability.
        rng (numpy.random.Generator): Random generator.

    returns:
        numpy.ndarray: Mutated and mixed individual.

    """
    j = rng.integers(len(pop[ic]))
    p = [1 / (len(pop) - 1.0) if i != ic else 0 for i in range(len(pop))] if len(pop) > 2 else None
    r = rng.choice(len(pop), 2, replace=not len(pop) >= 2, p=p)
    x = [x_b[i] + f * (pop[r[0]][i] - pop[r[1]][i]) if rng.random() < cr or i == j else pop[ic][i] for i in
         range(len(pop[ic]))]
    return np.asarray(x)


def cross_rand2(pop, ic, f, cr, rng, **_kwargs):
    r"""Mutation strategy with crossover.

    Mutation strategy uses five different random individuals from population.

    Mutation:
        Name: de/best/1

        :math:`\mathbf{v}_{i, G} = \mathbf{x}_{r_1, G} + differential_weight \cdot (\mathbf{x}_{r_2, G} - \mathbf{x}_{r_3, G}) + differential_weight \cdot (\mathbf{x}_{r_4, G} - \mathbf{x}_{r_5, G})`
        where :math:`r_1, r_2, r_3, r_4, r_5` are random indexes representing current population individuals.

    Crossover:
        Name: Binomial crossover

        :math:`\mathbf{x}_{i, G+1} = \begin{cases} \mathbf{u}_{i, G+1}, & \text{if $f(\mathbf{u}_{i, G+1}) \leq f(\mathbf{x}_{i, G})$}, \\ \mathbf{x}_{i, G}, & \text{otherwise}. \end{cases}`

    Args:
        pop (numpy.ndarray[Individual]): Current population.
        ic (int): Index of individual being mutated.
        f (float): Scale factor.
        cr (float): Crossover probability.
        rng (numpy.random.Generator): Random generator.

    Returns:
        numpy.ndarray: mutated and mixed individual.

    """
    j = rng.integers(len(pop[ic]))
    p = [1 / (len(pop) - 1.0) if i != ic else 0 for i in range(len(pop))] if len(pop) > 5 else None
    r = rng.choice(len(pop), 5, replace=not len(pop) >= 5, p=p)
    x = [pop[r[0]][i] + f * (pop[r[1]][i] - pop[r[2]][i]) + f * (
                pop[r[3]][i] - pop[r[4]][i]) if rng.random() < cr or i == j else pop[ic][i] for i in
         range(len(pop[ic]))]
    return np.asarray(x)


def cross_best2(pop, ic, f, cr, rng, x_b=None, **_kwargs):
    r"""Mutation strategy with crossover.

    Mutation:
        Name: de/best/2

        :math:`\mathbf{v}_{i, G} = \mathbf{x}_{best, G} + differential_weight \cdot (\mathbf{x}_{r_1, G} - \mathbf{x}_{r_2, G}) + differential_weight \cdot (\mathbf{x}_{r_3, G} - \mathbf{x}_{r_4, G})`
        where :math:`r_1, r_2, r_3, r_4` are random indexes representing current population individuals.

    Crossover:
        Name: Binomial crossover

        :math:`\mathbf{x}_{i, G+1} = \begin{cases} \mathbf{u}_{i, G+1}, & \text{if $f(\mathbf{u}_{i, G+1}) \leq f(\mathbf{x}_{i, G})$}, \\ \mathbf{x}_{i, G}, & \text{otherwise}. \end{cases}`

    Args:
        pop (numpy.ndarray[Individual]): Current population.
        ic (int): Index of individual being mutated.
        x_b (Individual): Current global best individual.
        f (float): Scale factor.
        cr (float): Crossover probability.
        rng (numpy.random.Generator): Random generator.

    Returns:
        numpy.ndarray: mutated and mixed individual.

    """
    j = rng.integers(len(pop[ic]))
    p = [1 / (len(pop) - 1.0) if i != ic else 0 for i in range(len(pop))] if len(pop) > 4 else None
    r = rng.choice(len(pop), 4, replace=not len(pop) >= 4, p=p)
    x = [x_b[i] + f * (pop[r[0]][i] - pop[r[1]][i]) + f * (
                pop[r[2]][i] - pop[r[3]][i]) if rng.random() < cr or i == j else pop[ic][i] for i in
         range(len(pop[ic]))]
    return np.asarray(x)


def cross_curr2rand1(pop, ic, f, cr, rng, **_kwargs):
    r"""Mutation strategy with crossover.

    Mutation:
        Name: de/curr2rand/1

        :math:`\mathbf{v}_{i, G} = \mathbf{x}_{i, G} + differential_weight \cdot (\mathbf{x}_{r_1, G} - \mathbf{x}_{r_2, G}) + differential_weight \cdot (\mathbf{x}_{r_3, G} - \mathbf{x}_{r_4, G})`
        where :math:`r_1, r_2, r_3, r_4` are random indexes representing current population individuals

    Crossover:
        Name: Binomial crossover

        :math:`\mathbf{x}_{i, G+1} = \begin{cases} \mathbf{u}_{i, G+1}, & \text{if $f(\mathbf{u}_{i, G+1}) \leq f(\mathbf{x}_{i, G})$}, \\ \mathbf{x}_{i, G}, & \text{otherwise}. \end{cases}`

    Args:
        pop (numpy.ndarray[Individual]): Current population.
        ic (int): Index of individual being mutated.
        f (float): Scale factor.
        cr (float): Crossover probability.
        rng (numpy.random.Generator): Random generator.

    Returns:
        numpy.ndarray: mutated and mixed individual.

    """
    j = rng.integers(len(pop[ic]))
    p = [1 / (len(pop) - 1.0) if i != ic else 0 for i in range(len(pop))] if len(pop) > 4 else None
    r = rng.choice(len(pop), 4, replace=not len(pop) >= 4, p=p)
    x = [pop[ic][i] + f * (pop[r[0]][i] - pop[r[1]][i]) + f * (
                pop[r[2]][i] - pop[r[3]][i]) if rng.random() < cr or i == j else pop[ic][i] for i in
         range(len(pop[ic]))]
    return np.asarray(x)


def cross_curr2best1(pop, ic, f, cr, rng, x_b=None, **_kwargs):
    r"""Mutation strategy with crossover.

    Mutation:
        Name: de/curr-to-best/1

        :math:`\mathbf{v}_{i, G} = \mathbf{x}_{i, G} + differential_weight \cdot (\mathbf{x}_{r_1, G} - \mathbf{x}_{r_2, G}) + differential_weight \cdot (\mathbf{x}_{r_3, G} - \mathbf{x}_{r_4, G})`
        where :math:`r_1, r_2, r_3, r_4` are random indexes representing current population individuals

    Crossover:
        Name: Binomial crossover

        :math:`\mathbf{x}_{i, G+1} = \begin{cases} \mathbf{u}_{i, G+1}, & \text{if $f(\mathbf{u}_{i, G+1}) \leq f(\mathbf{x}_{i, G})$}, \\ \mathbf{x}_{i, G}, & \text{otherwise}. \end{cases}`

    Args:
        pop (numpy.ndarray[Individual]): Current population.
        ic (int): Index of individual being mutated.
        x_b (Individual): Current global best individual.
        f (float): Scale factor.
        cr (float): Crossover probability.
        rng (numpy.random.Generator): Random generator.

    Returns:
        numpy.ndarray: mutated and mixed individual.

    """
    j = rng.integers(len(pop[ic]))
    p = [1 / (len(pop) - 1.0) if i != ic else 0 for i in range(len(pop))] if len(pop) > 3 else None
    r = rng.choice(len(pop), 3, replace=not len(pop) >= 3, p=p)
    x = [
        pop[ic][i] + f * (x_b[i] - pop[r[0]][i]) + f * (pop[r[1]][i] - pop[r[2]][i]) if rng.random() < cr or i == j else
        pop[ic][i] for i in range(len(pop[ic]))]
    return np.asarray(x)


class DifferentialEvolution(Algorithm):
    r"""Implementation of Differential evolution algorithm.

    Algorithm:
         Differential evolution algorithm

    Date:
        2018

    Author:
        Uros Mlakar and Klemen Berkovič

    License:
        MIT

    Reference paper:
        Storn, Rainer, and Kenneth Price. "Differential evolution - a simple and efficient heuristic for global optimization over continuous spaces." Journal of global optimization 11.4 (1997): 341-359.

    Attributes:
        Name (List[str]): List of string of names for algorithm.
        differential_weight (float): Scale factor.
        crossover_probability (float): Crossover probability.
        strategy (Callable[numpy.ndarray, int, numpy.ndarray, float, float, numpy.random.Generator, Dict[str, Any]]): crossover and mutation strategy.

    See Also:
        * :class:`niapy.algorithms.Algorithm`

    """

    Name = ['DifferentialEvolution', 'DE']

    @staticmethod
    def info():
        r"""Get basic information of algorithm.

        Returns:
            str: Basic information of algorithm.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""Storn, Rainer, and Kenneth Price. "Differential evolution - a simple and efficient heuristic for global optimization over continuous spaces." Journal of global optimization 11.4 (1997): 341-359."""

    def __init__(self, population_size=50, differential_weight=1, crossover_probability=0.8, strategy=cross_rand1,
                 *args, **kwargs):
        """Initialize DifferentialEvolution.

        Args:
            population_size (Optional[int]): Population size.
            differential_weight (Optional[float]): Differential weight (differential_weight).
            crossover_probability (Optional[float]): Crossover rate.
            strategy (Optional[Callable[[numpy.ndarray, int, numpy.ndarray, float, float, numpy.random.Generator, list], numpy.ndarray]]):
                Crossover and mutation strategy.

        See Also:
            * :func:`niapy.algorithms.Algorithm.__init__`

        """
        super().__init__(population_size,
                         initialization_function=kwargs.pop('initialization_function', default_individual_init),
                         individual_type=kwargs.pop('individual_type', Individual), *args, **kwargs)
        self.differential_weight = differential_weight
        self.crossover_probability = crossover_probability
        self.strategy = strategy

    def set_parameters(self, population_size=50, differential_weight=1, crossover_probability=0.8, strategy=cross_rand1,
                       **kwargs):
        r"""Set the algorithm parameters.

        Args:
            population_size (Optional[int]): Population size.
            differential_weight (Optional[float]): Differential weight (differential_weight).
            crossover_probability (Optional[float]): Crossover rate.
            strategy (Optional[Callable[[numpy.ndarray, int, numpy.ndarray, float, float, numpy.random.Generator, list], numpy.ndarray]]):
                Crossover and mutation strategy.

        See Also:
            * :func:`niapy.algorithms.Algorithm.set_parameters`

        """
        super().set_parameters(population_size=population_size,
                               initialization_function=kwargs.pop('initialization_function', default_individual_init),
                               individual_type=kwargs.pop('individual_type', Individual), **kwargs)
        self.differential_weight = differential_weight
        self.crossover_probability = crossover_probability
        self.strategy = strategy

    def get_parameters(self):
        r"""Get parameters values of the algorithm.

        Returns:
            Dict[str, Any]: Algorithm parameters.

        See Also:
            * :func:`niapy.algorithms.Algorithm.get_parameters`

        """
        d = Algorithm.get_parameters(self)
        d.update({
            'differential_weight': self.differential_weight,
            'crossover_probability': self.crossover_probability,
            'strategy': self.strategy
        })
        return d

    def evolve(self, pop, xb, task, **kwargs):
        r"""Evolve population.

        Args:
            pop (numpy.ndarray): Current population.
            xb (numpy.ndarray): Current best individual.
            task (Task): Optimization task.

        Returns:
            numpy.ndarray: New evolved populations.

        """
        return objects_to_array(
            [self.individual_type(x=self.strategy(pop, i, self.differential_weight, self.crossover_probability, self.rng, x_b=xb), task=task, rng=self.rng, e=True) for i
             in range(len(pop))])

    def selection(self, population, new_population, best_x, best_fitness, task, **kwargs):
        r"""Operator for selection.

        Args:
            population (numpy.ndarray): Current population.
            new_population (numpy.ndarray): New Population.
            best_x (numpy.ndarray): Current global best solution.
            best_fitness (float): Current global best solutions fitness/objective value.
            task (Task): Optimization task.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, float]:
                1. New selected individuals.
                2. New global best solution.
                3. New global best solutions fitness/objective value.

        """
        arr = objects_to_array([e if e.f < population[i].f else population[i] for i, e in enumerate(new_population)])
        best_x, best_fitness = self.get_best(arr, np.asarray([e.f for e in arr]), best_x, best_fitness)
        return arr, best_x, best_fitness

    def post_selection(self, pop, task, xb, fxb, **kwargs):
        r"""Apply additional operation after selection.

        Args:
            pop (numpy.ndarray): Current population.
            task (Task): Optimization task.
            xb (numpy.ndarray): Global best solution.
            fxb (float): Global best fitness.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, float]:
                1. New population.
                2. New global best solution.
                3. New global best solutions fitness/objective value.

        """
        return pop, xb, fxb

    def run_iteration(self, task, population, population_fitness, best_x, best_fitness, **params):
        r"""Core function of Differential Evolution algorithm.

        Args:
            task (Task): Optimization task.
            population (numpy.ndarray): Current population.
            population_fitness (numpy.ndarray): Current populations fitness/function values.
            best_x (numpy.ndarray): Current best individual.
            best_fitness (float): Current best individual function/fitness value.
            **params (dict): Additional arguments.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, float, Dict[str, Any]]:
                1. New population.
                2. New population fitness/function values.
                3. New global best solution.
                4. New global best solutions fitness/objective value.
                5. Additional arguments.

        See Also:
            * :func:`niapy.algorithms.basic.DifferentialEvolution.evolve`
            * :func:`niapy.algorithms.basic.DifferentialEvolution.selection`
            * :func:`niapy.algorithms.basic.DifferentialEvolution.post_selection`

        """
        new_population = self.evolve(population, best_x, task)
        population, best_x, best_fitness = self.selection(population, new_population, best_x, best_fitness, task=task)
        population, best_x, best_fitness = self.post_selection(population, task, best_x, best_fitness)
        population_fitness = np.asarray([x.f for x in population])
        best_x, best_fitness = self.get_best(population, population_fitness, best_x, best_fitness)
        return population, population_fitness, best_x, best_fitness, {}


class DynNpDifferentialEvolution(DifferentialEvolution):
    r"""Implementation of Dynamic population size Differential evolution algorithm.

    Algorithm:
        Dynamic population size Differential evolution algorithm

    Date:
        2018

    Author:
        Klemen Berkovič

    License:
        MIT

    Attributes:
        Name (List[str]): List of strings representing algorithm names.
        p_max (int): Number of population reductions.
        rp (int): Small non-negative number which is added to value of generations.

    See Also:
        * :class:`niapy.algorithms.basic.DifferentialEvolution`

    """

    Name = ['DynNpDifferentialEvolution', 'dynNpDE']

    @staticmethod
    def info():
        r"""Get basic information of algorithm.

        Returns:
            str: Basic information of algorithm.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""No info"""

    def __init__(self, population_size=10, p_max=50, rp=3, *args, **kwargs):
        """Initialize DynNpDifferentialEvolution.

        Args:
            p_max (Optional[int]): Number of population reductions.
            rp (Optional[int]): Small non-negative number which is added to value of generations.

        See Also:
            * :func:`niapy.algorithms.basic.DifferentialEvolution.__init__`

        """
        super().__init__(population_size, *args, **kwargs)
        self.p_max = p_max
        self.rp = rp

    def set_parameters(self, p_max=50, rp=3, **kwargs):
        r"""Set the algorithm parameters.

        Args:
            p_max (Optional[int]): Number of population reductions.
            rp (Optional[int]): Small non-negative number which is added to value of generations.

        See Also:
            * :func:`niapy.algorithms.basic.DifferentialEvolution.set_parameters`

        """
        super().set_parameters(**kwargs)
        self.p_max = p_max
        self.rp = rp

    def post_selection(self, pop, task, xb, fxb, **kwargs):
        r"""Post selection operator.

        In this algorithm the post selection operator decrements the population at specific iterations/generations.

        Args:
            pop (numpy.ndarray): Current population.
            task (Task): Optimization task.
            xb (numpy.ndarray): Global best individual coordinates.
            fxb (float): Global best fitness.
            kwargs (Dict[str, Any]): Additional arguments.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, float]:
                1. Changed current population.
                2. New global best solution.
                3. New global best solutions fitness/objective value.

        """
        gr = task.max_evals // (self.p_max * len(pop)) + self.rp
        new_np = len(pop) // 2
        if (task.iters + 1) == gr and len(pop) > 3:
            pop = objects_to_array([pop[i] if pop[i].f < pop[i + new_np].f else pop[i + new_np] for i in range(new_np)])
        return pop, xb, fxb


def proportional(min_lifetime, max_lifetime, mu, x_f, avg, **_kwargs):
    r"""Proportional calculation of age of individual.

    Args:
        min_lifetime (int): Minimal life time.
        max_lifetime (int): Maximal life time.
        mu (float): Median of life time.
        x_f (float): Individuals function/fitness value.
        avg (float): Average fitness/function value of current population.

    Returns:
        int: Age of individual.

    """
    proportional_result = max_lifetime if math.isinf(avg) else min_lifetime + mu * avg / x_f
    return min(proportional_result, max_lifetime)


def linear(min_lifetime, mu, x_f, x_gw, x_gb, **_kwargs):
    r"""Linear calculation of age of individual.

    Args:
        min_lifetime (int): Minimal life time.
        mu (float): Median of life time.
        x_f (float): Individual function/fitness value.
        x_gw (float): Global worst fitness/function value.
        x_gb (float): Global best fitness/function value.

    Returns:
        int: Age of individual.

    """
    return min_lifetime + 2 * mu * (x_f - x_gw) / (x_gb - x_gw)


def bilinear(min_lifetime, max_lifetime, mu, x_f, avg, x_gw, x_gb, **_kwargs):
    r"""Bilinear calculation of age of individual.

    Args:
        min_lifetime (int): Minimal life time.
        max_lifetime (int): Maximal life time.
        mu (float): Median of life time.
        x_f (float): Individual function/fitness value.
        avg (float): Average fitness/function value.
        x_gw (float): Global worst fitness/function value.
        x_gb (float): Global best fitness/function value.

    Returns:
        int: Age of individual.

    """
    if avg < x_f:
        return min_lifetime + mu * (x_f - x_gw) / (x_gb - x_gw)
    return 0.5 * (min_lifetime + max_lifetime) + mu * (x_f - avg) / (x_gb - avg)


class AgingIndividual(Individual):
    r"""Individual with aging.

    Attributes:
        age (int): Age of individual.

    See Also:
        * :class:`niapy.algorithms.Individual`

    """

    def __init__(self, **kwargs):
        r"""Init Aging Individual.

        See Also:
            * :func:`niapy.algorithms.Individual.__init__`

        """
        super().__init__(**kwargs)
        self.age = 0


class AgingNpDifferentialEvolution(DifferentialEvolution):
    r"""Implementation of Differential evolution algorithm with aging individuals.

    Algorithm:
        Differential evolution algorithm with dynamic population size that is defined by the quality of population

    Date:
        2018

    Author:
        Klemen Berkovič

    License:
        MIT

    Attributes:
        Name (List[str]): list of strings representing algorithm names.
        Lt_min (int): Minimal age of individual.
        Lt_max (int): Maximal age of individual.
        delta_np (float): Proportion of how many individuals shall die.
        omega (float): Acceptance rate for individuals to die.
        mu (int): Mean of individual max and min age.
        age (Callable[[int, int, float, float, float, float, float], int]): Function for calculation of age for individual.

    See Also:
        * :class:`niapy.algorithms.basic.DifferentialEvolution`

    """

    Name = ['AgingNpDifferentialEvolution', 'ANpDE']

    @staticmethod
    def info():
        r"""Get basic information of algorithm.

        Returns:
            str: Basic information of algorithm.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""No info"""

    def __init__(self, min_lifetime=0, max_lifetime=12, delta_np=0.3, omega=0.3, age=proportional, *args, **kwargs):
        """Initialize AgingNpDifferentialEvolution.

        Arguments:
            min_lifetime (Optional[int]): Minimum life time.
            max_lifetime (Optional[int]): Maximum life time.
            delta_np (Optional[float]): Proportion of how many individuals shall die.
            omega (Optional[float]): Acceptance rate for individuals to die.
            age (Optional[Callable[[int, int, float, float, float, float, float], int]]): Function for calculation of age for individual.

        See Also:
            * :func:`niapy.algorithms.basic.DifferentialEvolution.__init__`

        """
        super().__init__(individual_type=AgingIndividual, *args, **kwargs)
        self.min_lifetime = min_lifetime
        self.max_lifetime = max_lifetime
        self.age = age
        self.delta_np = delta_np
        self.omega = omega
        self.mu = abs(self.max_lifetime - self.min_lifetime) / 2

    def set_parameters(self, min_lifetime=0, max_lifetime=12, delta_np=0.3, omega=0.3, age=proportional, **kwargs):
        r"""Set the algorithm parameters.

        Arguments:
            min_lifetime (Optional[int]): Minimum life time.
            max_lifetime (Optional[int]): Maximum life time.
            delta_np (Optional[float]): Proportion of how many individuals shall die.
            omega (Optional[float]): Acceptance rate for individuals to die.
            age (Optional[Callable[[int, int, float, float, float, float, float], int]]): Function for calculation of age for individual.

        See Also:
            * :func:`niapy.algorithms.basic.DifferentialEvolution.set_parameters`

        """
        super().set_parameters(individual_type=AgingIndividual, **kwargs)
        self.min_lifetime = min_lifetime
        self.max_lifetime = max_lifetime
        self.age = age
        self.delta_np = delta_np
        self.omega = omega
        self.mu = abs(self.max_lifetime - self.min_lifetime) / 2

    def delta_pop_eliminated(self, t):
        r"""Calculate how many individuals are going to die.

        Args:
            t (int): Number of generations made by the algorithm.

        Returns:
            int: Number of individuals to dye.

        """
        return int(self.delta_np * abs(np.sin(t)))

    def delta_pop_created(self, t):
        r"""Calculate how many individuals are going to be created.

        Args:
            t (int): Number of generations made by the algorithm.

        Returns:
            int: Number of individuals to be born.

        """
        return int(self.delta_np * abs(np.cos(t)))

    def aging(self, task, pop):
        r"""Apply aging to individuals.

        Args:
            task (Task): Optimization task.
            pop (numpy.ndarray[Individual]): Current population.

        Returns:
            numpy.ndarray[Individual]: New population.

        """
        fpop = np.asarray([x.f for x in pop])
        x_b, x_w = pop[np.argmin(fpop)], pop[np.argmax(fpop)]
        avg = np.mean(fpop[np.isfinite(fpop)])
        new_population = []
        for x in pop:
            x.age += 1
            lifetime = round(
                self.age(min_lifetime=self.min_lifetime, max_lifetime=self.max_lifetime, mu=self.mu, x_f=x.f, avg=avg, x_gw=x_w.f, x_gb=x_b.f))
            if x.age <= lifetime:
                new_population.append(x)
        if len(new_population) == 0:
            new_population = objects_to_array([self.individual_type(task=task, rng=self.rng, e=True) for _ in range(self.population_size)])
        return new_population

    def increment_population(self, task):
        r"""Increment population.

        Args:
            task (Task): Optimization task.

        Returns:
            numpy.ndarray[Individual]: Increased population.

        """
        delta_pop = int(round(max(1, self.population_size * self.delta_pop_eliminated((task.iters + 1)))))
        return objects_to_array([self.individual_type(task=task, rng=self.rng, e=True) for _ in range(delta_pop)])

    def decrement_population(self, pop, task):
        r"""Decrement population.

        Args:
            pop (numpy.ndarray): Current population.
            task (Task): Optimization task.

        Returns:
            numpy.ndarray[Individual]: Decreased population.

        """
        delta_population = int(round(max(1, self.population_size * self.delta_pop_created((task.iters + 1)))))
        if len(pop) - delta_population <= 0:
            return pop
        ni = self.rng.choice(len(pop), delta_population, replace=False)
        new_population = []
        for i, e in enumerate(pop):
            if i not in ni:
                new_population.append(e)
            elif self.random() >= self.omega:
                new_population.append(e)
        return objects_to_array(new_population)

    def selection(self, population, new_population, best_x, best_fitness, task, **kwargs):
        r"""Select operator for individuals with aging.

        Args:
            population (numpy.ndarray): Current population.
            new_population (numpy.ndarray): New population.
            best_x (numpy.ndarray): Current global best solution.
            best_fitness (float): Current global best solutions fitness/objective value.
            task (Task): Optimization task.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, float]:
                1. New population of individuals.
                2. New global best solution.
                3. New global best solutions fitness/objective value.

        """
        new_population, best_x, best_fitness = super().selection(population, new_population, best_x, best_fitness, task)
        new_population = np.append(new_population, self.increment_population(task))
        best_x, best_fitness = self.get_best(new_population, np.asarray([e.f for e in new_population]), best_x, best_fitness)
        population = self.aging(task, new_population)
        return population, best_x, best_fitness

    def post_selection(self, pop, task, xb, fxb, **kwargs):
        r"""Post selection operator.

        Args:
            pop (numpy.ndarray): Current population.
            task (Task): Optimization task.
            xb (Individual): Global best individual.
            fxb (float): Global best fitness.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, float]:
            1. New population.
            2. New global best solution
            3. New global best solutions fitness/objective value

        """
        return self.decrement_population(pop, task) if len(pop) > self.population_size else pop, xb, fxb


def multi_mutations(pop, i, xb, differential_weight, crossover_probability, rng, task, individual_type, strategies,
                    **_kwargs):
    r"""Mutation strategy that takes more than one strategy and applies them to individual.

    Args:
        pop (numpy.ndarray[Individual]): Current population.
        i (int): Index of current individual.
        xb (Individual): Current best individual.
        differential_weight (float): Scale factor.
        crossover_probability (float): Crossover probability.
        rng (numpy.random.Generator): Random generator.
        task (Task): Optimization task.
        individual_type (Type[Individual]): Individual type used in algorithm.
        strategies (Iterable[Callable[[numpy.ndarray[Individual], int, Individual, float, float, numpy.random.Generator], numpy.ndarray[Individual]]]): List of mutation strategies.

    Returns:
        Individual: Best individual from applied mutations strategies.

    """
    population = [individual_type(x=strategy(pop, i, differential_weight, crossover_probability, x_b=xb, rng=rng), task=task, e=True, rng=rng) for strategy in strategies]
    return population[np.argmin([x.f for x in population])]


class MultiStrategyDifferentialEvolution(DifferentialEvolution):
    r"""Implementation of Differential evolution algorithm with multiple mutation strategies.

    Algorithm:
        Implementation of Differential evolution algorithm with multiple mutation strategies

    Date:
        2018

    Author:
        Klemen Berkovič

    License:
        MIT

    Attributes:
        Name (List[str]): List of strings representing algorithm names.
        strategies (Iterable[Callable[[numpy.ndarray[Individual], int, Individual, float, float, numpy.random.Generator], numpy.ndarray[Individual]]]): List of mutation strategies.

    See Also:
        * :class:`niapy.algorithms.basic.DifferentialEvolution`

    """

    Name = ['MultiStrategyDifferentialEvolution', 'MsDE']

    @staticmethod
    def info():
        r"""Get basic information of algorithm.

        Returns:
            str: Basic information of algorithm.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""No info"""

    def __init__(self, population_size=40, strategies=(cross_rand1, cross_best1, cross_curr2best1, cross_rand2), *args, **kwargs):
        """Initialize MultiStrategyDifferentialEvolution.

        Args:
            strategies (Optional[Iterable[Callable[[numpy.ndarray[Individual], int, Individual, float, float, numpy.random.Generator], numpy.ndarray[Individual]]]]):
                List of mutation strategies.

        See Also:
            * :func:`niapy.algorithms.basic.DifferentialEvolution.__init__`

        """
        super().__init__(population_size, strategy=multi_mutations, *args, **kwargs)
        self.strategies = strategies

    def set_parameters(self, strategies=(cross_rand1, cross_best1, cross_curr2best1, cross_rand2), **kwargs):
        r"""Set the arguments of the algorithm.

        Args:
            strategies (Optional[Iterable[Callable[[numpy.ndarray[Individual], int, Individual, float, float, numpy.random.Generator], numpy.ndarray[Individual]]]]):
                List of mutation strategies.

        See Also:
            * :func:`niapy.algorithms.basic.DifferentialEvolution.set_parameters`

        """
        super().set_parameters(strategy=multi_mutations, **kwargs)
        self.strategies = strategies

    def get_parameters(self):
        r"""Get parameters values of the algorithm.

        Returns:
            Dict[str, Any]: Algorithm parameters.

        See Also:
            * :func:`niapy.algorithms.basic.DifferentialEvolution.get_parameters`

        """
        d = DifferentialEvolution.get_parameters(self)
        d.update({'strategies': self.strategies})
        return d

    def evolve(self, pop, xb, task, **kwargs):
        r"""Evolve population with the help multiple mutation strategies.

        Args:
            pop (numpy.ndarray): Current population.
            xb (numpy.ndarray): Current best individual.
            task (Task): Optimization task.

        Returns:
            numpy.ndarray: New population of individuals.

        """
        return objects_to_array(
            [self.strategy(pop, i, xb, self.differential_weight, self.crossover_probability, self.rng, task, self.individual_type, self.strategies) for i in
             range(len(pop))])


class DynNpMultiStrategyDifferentialEvolution(MultiStrategyDifferentialEvolution, DynNpDifferentialEvolution):
    r"""Implementation of Dynamic population size Differential evolution algorithm with dynamic population size that is defined by the quality of population.

    Algorithm:
        Dynamic population size Differential evolution algorithm with dynamic population size that is defined by the quality of population

    Date:
        2018

    Author:
        Klemen Berkovič

    License:
        MIT

    Attributes:
        Name (List[str]): List of strings representing algorithm name.

    See Also:
        * :class:`niapy.algorithms.basic.MultiStrategyDifferentialEvolution`
        * :class:`niapy.algorithms.basic.DynNpDifferentialEvolution`

    """

    Name = ['DynNpMultiStrategyDifferentialEvolution', 'dynNpMsDE']

    @staticmethod
    def info():
        r"""Get basic information of algorithm.

        Returns:
            str: Basic information of algorithm.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""No info"""

    def set_parameters(self, **kwargs):
        r"""Set the arguments of the algorithm.

        See Also:
            * :func:`niapy.algorithms.basic.MultiStrategyDifferentialEvolution.set_parameters`
            * :func:`niapy.algorithms.basic.DynNpDifferentialEvolution.set_parameters`

        """
        DynNpDifferentialEvolution.set_parameters(self, **kwargs)
        MultiStrategyDifferentialEvolution.set_parameters(self, **kwargs)

    def evolve(self, pop, xb, task, **kwargs):
        r"""Evolve the current population.

        Args:
            pop (numpy.ndarray): Current population.
            xb (numpy.ndarray): Global best solution.
            task (Task): Optimization task.

        Returns:
            numpy.ndarray: Evolved new population.

        """
        return MultiStrategyDifferentialEvolution.evolve(self, pop, xb, task, **kwargs)

    def post_selection(self, pop, task, xb, fxb, **kwargs):
        r"""Post selection operator.

        Args:
            pop (numpy.ndarray): Current population.
            task (Task): Optimization task.
            xb (numpy.ndarray): Global best individual
            fxb (float): Global best fitness.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, float]:
                1. New population.
                2. New global best solution.
                3. New global best solutions fitness/objective value.

        See Also:
            * :func:`niapy.algorithms.basic.DynNpDifferentialEvolution.post_selection`

        """
        return DynNpDifferentialEvolution.post_selection(self, pop, task, xb, fxb)



# encoding=utf8

"""Implementation of Ridge function."""

import numpy as np
from niapy.problems.problem import Problem

__all__ = ['Ridge']


class Ridge(Problem):
    r"""Implementation of Ridge function.

    Date: 2018

    Author: Lucija Brezočnik

    License: MIT

    Function: **Ridge function**

        :math:`f(\mathbf{x}) = \sum_{i=1}^D (\sum_{j=1}^i x_j)^2`

        **Input domain:**
        The function can be defined on any input domain but it is usually
        evaluated on the hypercube :math:`x_i ∈ [-64, 64]`, for all :math:`i = 1, 2,..., D`.

        **Global minimum:** :math:`f(x^*) = 0`, at :math:`x^* = (0,...,0)`

    LaTeX formats:
        Inline:
            $f(\mathbf{x}) = \sum_{i=1}^D (\sum_{j=1}^i x_j)^2 $

        Equation:
            \begin{equation} f(\mathbf{x}) =
            \sum_{i=1}^D (\sum_{j=1}^i x_j)^2 \end{equation}

        Domain:
            $-64 \leq x_i \leq 64$

    Reference:
        http://www.cs.unm.edu/~neal.holts/dga/benchmarkFunction/ridge.html

    """

    def __init__(self, dimension=4, lower=-64.0, upper=64.0, *args, **kwargs):
        r"""Initialize Ridge problem..

        Args:
            dimension (Optional[int]): Dimension of the problem.
            lower (Optional[Union[float, Iterable[float]]]): Lower bounds of the problem.
            upper (Optional[Union[float, Iterable[float]]]): Upper bounds of the problem.

        See Also:
            :func:`niapy.problems.Problem.__init__`

        """
        super().__init__(dimension, lower, upper, *args, **kwargs)

    @staticmethod
    def latex_code():
        r"""Return the latex code of the problem.

        Returns:
            str: Latex code.

        """
        return r'''$f(\mathbf{x}) = \sum_{i=1}^D (\sum_{j=1}^i x_j)^2 $'''

    def _evaluate(self, x):
        inner = np.array([np.sum(x[:i]) for i in range(1, self.dimension + 1)])
        return np.sum(inner ** 2)



import numpy as np

__all__ = ['full_array', 'objects_to_array']


def full_array(a, dimension):
    r"""Fill or create array of length dimension, from value or value form a.

    Args:
        a (Union[int, float, numpy.ndarray, Iterable[Any]]): Input values for fill.
        dimension (int): Length of new array.

    Returns:
        numpy.ndarray: Array filled with passed values or value.

    """
    if isinstance(a, (int, float)):
        out = np.ones(dimension) * a
    elif isinstance(a, (np.ndarray, list, tuple)):
        if len(a) == dimension:
            out = np.asarray(a)
        elif len(a) > dimension:
            out = np.asarray(a[:dimension])
        else:
            out = np.tile(a, int(np.ceil(dimension / len(a))))[:dimension]
    else:
        raise TypeError('`a` must be a scalar or an Iterable.')
    return out


def objects_to_array(objs):
    r"""Convert `Iterable` array or list to `NumPy` array with dtype object.

    Args:
        objs (Iterable[Any]): Array or list to convert.

    Returns:
        numpy.ndarray: Array of objects.

    """
    a = np.empty(len(objs), dtype=object)
    for i, e in enumerate(objs):
        a[i] = e
    return a



# encoding=utf8
import logging
import multiprocessing
import threading

import numpy as np
from numpy.random import default_rng

from niapy.util.array import objects_to_array

logging.basicConfig()
logger = logging.getLogger('niapy.util.utility')
logger.setLevel('INFO')

__all__ = [
    'Algorithm',
    'Individual',
    'default_individual_init',
    'default_numpy_init'
]


def default_numpy_init(task, population_size, rng, **_kwargs):
    r"""Initialize starting population that is represented with `numpy.ndarray` with shape `(population_size, task.dimension)`.

    Args:
        task (Task): Optimization task.
        population_size (int): Number of individuals in population.
        rng (numpy.random.Generator): Random number generator.

    Returns:
        Tuple[numpy.ndarray, numpy.ndarray[float]]:
            1. New population with shape `(population_size, task.D)`.
            2. New population function/fitness values.

    """
    pop = rng.uniform(task.lower, task.upper, (population_size, task.dimension))
    fpop = np.apply_along_axis(task.eval, 1, pop)
    return pop, fpop


def default_individual_init(task, population_size, rng, individual_type=None, **_kwargs):
    r"""Initialize `population_size` individuals of type `individual_type`.

    Args:
        task (Task): Optimization task.
        population_size (int): Number of individuals in population.
        rng (numpy.random.Generator): Random number generator.
        individual_type (Optional[Individual]): Class of individual in population.

    Returns:
        Tuple[numpy.ndarray[Individual], numpy.ndarray[float]:
            1. Initialized individuals.
            2. Initialized individuals function/fitness values.

    """
    pop = objects_to_array([individual_type(task=task, rng=rng, e=True) for _ in range(population_size)])
    return pop, np.asarray([x.f for x in pop])


class Algorithm:
    r"""Class for implementing algorithms.

    Date:
        2018

    Author
        Klemen Berkovič

    License:
        MIT

    Attributes:
        Name (List[str]): List of names for algorithm.
        rng (numpy.random.Generator): Random generator.
        population_size (int): Population size.
        initialization_function (Callable[[int, Task, numpy.random.Generator, Dict[str, Any]], Tuple[numpy.ndarray, numpy.ndarray[float]]]):
            Population initialization function.
        individual_type (Optional[Type[Individual]]): Type of individuals used in population, default value is None for Numpy arrays.

    """

    Name = ['Algorithm', 'AAA']

    def __init__(self, population_size=50, initialization_function=default_numpy_init, individual_type=None,
                 seed=None, *args, **kwargs):
        r"""Initialize algorithm and create name for an algorithm.

        Args:
            population_size (Optional[int]): Population size.
            initialization_function (Optional[Callable[[int, Task, numpy.random.Generator, Dict[str, Any]], Tuple[numpy.ndarray, numpy.ndarray[float]]]]):
                Population initialization function.
            individual_type (Optional[Type[Individual]]): Individual type used in population, default is Numpy array.
            seed (Optional[int]): Starting seed for random generator.

        See Also:
            * :func:`niapy.algorithms.Algorithm.set_parameters`

        """
        self.population_size = population_size
        self.initialization_function = initialization_function
        self.individual_type = individual_type
        self.rng = default_rng(seed)
        self.exception = None

    @staticmethod
    def info():
        r"""Get algorithm information.

        Returns:
            str: Bit item.

        """
        return '''Basic algorithm. No implementation!!!'''

    def set_parameters(self, population_size=50, initialization_function=default_numpy_init, individual_type=None,
                       *args, **kwargs):
        r"""Set the parameters/arguments of the algorithm.

        Args:
            population_size (Optional[int]): Population size.
            initialization_function (Optional[Callable[[int, Task, numpy.random.Generator, Dict[str, Any]], Tuple[numpy.ndarray, numpy.ndarray[float]]]]):
                Population initialization function.
            individual_type (Optional[Type[Individual]]): Individual type used in population, default is Numpy array.

        See Also:
            * :func:`niapy.algorithms.default_numpy_init`
            * :func:`niapy.algorithms.default_individual_init`

        """
        self.population_size = population_size
        self.initialization_function = initialization_function
        self.individual_type = individual_type

    def get_parameters(self):
        r"""Get parameters of the algorithm.

        Returns:
            Dict[str, Any]:
                * Parameter name (str): Represents a parameter name
                * Value of parameter (Any): Represents the value of the parameter

        """
        return {
            'population_size': self.population_size,
            'initialization_function': self.initialization_function,
            'individual_type': self.individual_type
        }

    def random(self, size=None):
        r"""Get random distribution of shape size in range from 0 to 1.

        Args:
            size (Union[None, int, Iterable[int]]): Shape of returned random distribution.

        Returns:
            Union[numpy.ndarray[float], float]: Random number or numbers :math:`\in [0, 1]`.

        """
        return self.rng.random(size)

    def uniform(self, low, high, size=None):
        r"""Get uniform random distribution of shape size in range from "low" to "high".

        Args:
            low (Union[float, Iterable[float]]): Lower bound.
            high (Union[float, Iterable[float]]): Upper bound.
            size (Union[None, int, Iterable[int]]): Shape of returned uniform random distribution.

        Returns:
            Union[numpy.ndarray[float], float]: Array of numbers :math:`\in [\mathit{Lower}, \mathit{Upper}]`.

        """
        return self.rng.uniform(low, high, size)

    def normal(self, loc, scale, size=None):
        r"""Get normal random distribution of shape size with mean "loc" and standard deviation "scale".

        Args:
            loc (float): Mean of the normal random distribution.
            scale (float): Standard deviation of the normal random distribution.
            size (Union[int, Iterable[int]]): Shape of returned normal random distribution.

        Returns:
            Union[numpy.ndarray[float], float]: Array of numbers.

        """
        return self.rng.normal(loc, scale, size)

    def standard_normal(self, size=None):
        r"""Get standard normal distribution of shape size.

        Args:
            size (Union[int, Iterable[int]]): Shape of returned standard normal distribution.

        Returns:
            Union[numpy.ndarray[float], float]: Random generated numbers or one random generated number :math:`\in [0, 1]`.

        """
        return self.rng.standard_normal(size)

    def integers(self, low, high=None, size=None, skip=None):
        r"""Get discrete uniform (integer) random distribution of D shape in range from "low" to "high".

        Args:
            low (Union[int, Iterable[int]]): Lower integer bound.
                If high = None low is 0 and this value is used as high
            high (Union[int, Iterable[int]]): One above upper integer bound.
            size (Union[None, int, Iterable[int]]): shape of returned discrete uniform random distribution.
            skip (Union[None, int, Iterable[int], numpy.ndarray[int]]): numbers to skip.

        Returns:
            Union[int, numpy.ndarray[int]]: Random generated integer number.

        """
        r = self.rng.integers(low, high, size)
        return r if skip is None or r not in skip else self.integers(low, high, size, skip)

    @staticmethod
    def get_best(population, population_fitness, best_x=None, best_fitness=np.inf):
        r"""Get the best individual for population.

        Args:
            population (numpy.ndarray): Current population.
            population_fitness (numpy.ndarray): Current populations fitness/function values of aligned individuals.
            best_x (Optional[numpy.ndarray]): Best individual.
            best_fitness (float): Fitness value of best individual.

        Returns:
            Tuple[numpy.ndarray, float]:
                1. Coordinates of best solution.
                2. beset fitness/function value.

        """
        ib = np.argmin(population_fitness)
        if isinstance(population_fitness, (float, int)) and best_fitness >= population_fitness:
            best_x, best_fitness = population, population_fitness
        elif isinstance(population_fitness, (np.ndarray, list)) and best_fitness >= population_fitness[ib]:
            best_x, best_fitness = population[ib], population_fitness[ib]
        return (best_x.x.copy() if isinstance(best_x, Individual) else best_x.copy()), best_fitness

    def init_population(self, task):
        r"""Initialize starting population of optimization algorithm.

        Args:
            task (Task): Optimization task.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, Dict[str, Any]]:
                1. New population.
                2. New population fitness values.
                3. Additional arguments.

        See Also:
            * :func:`niapy.algorithms.Algorithm.set_parameters`

        """
        pop, fpop = self.initialization_function(task=task, population_size=self.population_size, rng=self.rng,
                                                 individual_type=self.individual_type)
        return pop, fpop, {}

    def run_iteration(self, task, population, population_fitness, best_x, best_fitness, **params):
        r"""Core functionality of algorithm.

        This function is called on every algorithm iteration.

        Args:
            task (Task): Optimization task.
            population (numpy.ndarray): Current population coordinates.
            population_fitness (numpy.ndarray): Current population fitness value.
            best_x (numpy.ndarray): Current generation best individuals coordinates.
            best_fitness (float): current generation best individuals fitness value.
            **params (Dict[str, Any]): Additional arguments for algorithms.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, float, Dict[str, Any]]:
                1. New populations coordinates.
                2. New populations fitness values.
                3. New global best position/solution
                4. New global best fitness/objective value
                5. Additional arguments of the algorithm.

        See Also:
            * :func:`niapy.algorithms.Algorithm.iteration_generator`

        """
        return population, population_fitness, best_x, best_fitness, params

    def iteration_generator(self, task):
        r"""Run the algorithm for a single iteration and return the best solution.

        Args:
            task (Task): Task with bounds and objective function for optimization.

        Returns:
            Generator[Tuple[numpy.ndarray, float], None, None]: Generator getting new/old optimal global values.

        Yields:
            Tuple[numpy.ndarray, float]:
                1. New population best individuals coordinates.
                2. Fitness value of the best solution.

        See Also:
            * :func:`niapy.algorithms.Algorithm.init_population`
            * :func:`niapy.algorithms.Algorithm.run_iteration`

        """
        pop, fpop, params = self.init_population(task)
        xb, fxb = self.get_best(pop, fpop)
        if task.stopping_condition():
            yield xb, fxb
        while True:
            pop, fpop, xb, fxb, params = self.run_iteration(task, pop, fpop, xb, fxb, **params)
            yield xb, fxb

    def run_task(self, task):
        r"""Start the optimization.

        Args:
            task (Task): Task with bounds and objective function for optimization.

        Returns:
            Tuple[numpy.ndarray, float]:
                1. Best individuals components found in optimization process.
                2. Best fitness value found in optimization process.

        See Also:
            * :func:`niapy.algorithms.Algorithm.iteration_generator`

        """
        algo, xb, fxb = self.iteration_generator(task), None, np.inf
        while not task.stopping_condition():
            xb, fxb = next(algo)
            task.next_iter()
        return xb, fxb

    def run(self, task):
        r"""Start the optimization.

        Args:
            task (Task): Optimization task.

        Returns:
            Tuple[numpy.ndarray, float]:
                1. Best individuals components found in optimization process.
                2. Best fitness value found in optimization process.

        See Also:
            * :func:`niapy.algorithms.Algorithm.run_task`

        """
        try:
            r = self.run_task(task)
            return r[0], r[1] * task.optimization_type.value
        except BaseException as e:
            if threading.current_thread() == threading.main_thread() and multiprocessing.current_process().name == 'MainProcess':
                raise e
            self.exception = e
            return None, None

    def bad_run(self):
        r"""Check if some exceptions where thrown when the algorithm was running.

        Returns:
            bool: True if some error where detected at runtime of the algorithm, otherwise False

        """
        return self.exception is not None


class Individual:
    r"""Class that represents one solution in population of solutions.

    Date:
        2018

    Author:
        Klemen Berkovič

    License:
        MIT

    Attributes:
        x (numpy.ndarray): Coordinates of individual.
        f (float): Function/fitness value of individual.

    """

    def __init__(self, x=None, task=None, e=True, rng=None, **kwargs):
        r"""Initialize new individual.

        Parameters:
            task (Optional[Task]): Optimization task.
            rand (Optional[numpy.random.Generator]): Random generator.
            x (Optional[numpy.ndarray]): Individuals components.
            e (Optional[bool]): True to evaluate the individual on initialization. Default value is True.

        """
        self.f = task.optimization_type.value * np.inf if task is not None else np.inf
        if x is not None:
            self.x = x if isinstance(x, np.ndarray) else np.asarray(x)
        elif task is not None:
            self.generate_solution(task, default_rng(rng))
        if e and task is not None:
            self.evaluate(task, rng)

    def generate_solution(self, task, rng):
        r"""Generate new solution.

        Generate new solution for this individual and set it to ``self.x``.
        This method uses ``rng`` for getting random numbers.
        For generating random components ``rng`` and ``task`` is used.

        Args:
            task (Task): Optimization task.
            rng (numpy.random.Generator): Random numbers generator object.

        """
        self.x = rng.uniform(task.lower, task.upper, task.dimension)

    def evaluate(self, task, rng=None):
        r"""Evaluate the solution.

        Evaluate solution ``this.x`` with the help of task.
        Task is used for repairing the solution and then evaluating it.

        Args:
            task (Task): Objective function object.
            rng (Optional[numpy.random.Generator]): Random generator.

        See Also:
            * :func:`niapy.task.Task.repair`

        """
        self.x = task.repair(self.x, rng=rng)
        self.f = task.eval(self.x)

    def copy(self):
        r"""Return a copy of self.

        Method returns copy of ``this`` object so it is safe for editing.

        Returns:
            Individual: Copy of self.

        """
        return Individual(x=self.x.copy(), e=False, f=self.f)

    def __eq__(self, other):
        r"""Compare the individuals for equalities.

        Args:
            other (Union[Any, numpy.ndarray]): Object that we want to compare this object to.

        Returns:
            bool: `True` if equal or `False` if no equal.

        """
        if isinstance(other, np.ndarray):
            for e in other:
                if self == e:
                    return True
            return False
        return np.array_equal(self.x, other.x) and self.f == other.f

    def __str__(self):
        r"""Print the individual with the solution and objective value.

        Returns:
            str: String representation of self.

        """
        return '%s -> %s' % (self.x, self.f)

    def __getitem__(self, i):
        r"""Get the value of i-th component of the solution.

        Args:
            i (int): Position of the solution component.

        Returns:
            Any: Value of ith component.

        """
        return self.x[i]

    def __setitem__(self, i, v):
        r"""Set the value of i-th component of the solution to v value.

        Args:
            i (int): Position of the solution component.
            v (Any): Value to set to i-th component.

        """
        self.x[i] = v

    def __len__(self):
        r"""Get the length of the solution or the number of components.

        Returns:
            int: Number of components.

        """
        return len(self.x)



# encoding=utf8
import logging

from niapy.algorithms.basic.de import cross_best1
from niapy.algorithms.modified.saba import SelfAdaptiveBatAlgorithm

logging.basicConfig()
logger = logging.getLogger('niapy.algorithms.modified')
logger.setLevel('INFO')

__all__ = ['HybridSelfAdaptiveBatAlgorithm']


class HybridSelfAdaptiveBatAlgorithm(SelfAdaptiveBatAlgorithm):
    r"""Implementation of Hybrid self adaptive bat algorithm.

    Algorithm:
        Hybrid self adaptive bat algorithm

    Date:
        April 2019

    Author:
        Klemen Berkovič

    License:
        MIT

    Reference paper:
        Fister, Iztok, Simon Fong, and Janez Brest. "A novel hybrid self-adaptive bat algorithm." The Scientific World Journal 2014 (2014).

    Reference URL:
        https://www.hindawi.com/journals/tswj/2014/709738/cta/

    Attributes:
        Name (List[str]): List of strings representing algorithm name.
        F (float): Scaling factor for local search.
        CR (float): Probability of crossover for local search.
        CrossMutt (Callable[[numpy.ndarray, int, numpy.ndarray, float, float, numpy.random.Generator, Dict[str, Any]): Local search method based of Differential evolution strategy.

    See Also:
        * :class:`niapy.algorithms.basic.BatAlgorithm`

    """

    Name = ['HybridSelfAdaptiveBatAlgorithm', 'HSABA']

    @staticmethod
    def info():
        r"""Get basic information about the algorithm.

        Returns:
            str: Basic information.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""Fister, Iztok, Simon Fong, and Janez Brest. "A novel hybrid self-adaptive bat algorithm." The Scientific World Journal 2014 (2014)."""

    def __init__(self, differential_weight=0.9, crossover_probability=0.85, strategy=cross_best1, *args, **kwargs):
        """Initialize HybridSelfAdaptiveBatAlgorithm.

        Args:
            differential_weight (Optional[float]): Scaling factor for local search.
            crossover_probability (Optional[float]): Probability of crossover for local search.
            strategy (Optional[Callable[[numpy.ndarray, int, numpy.ndarray, float, float, numpy.random.Generator, Dict[str, Any], numpy.ndarray]]): Local search method based of Differential evolution strategy.

        See Also:
            * :func:`niapy.algorithms.basic.BatAlgorithm.__init__`

        """
        super().__init__(*args, **kwargs)
        self.differential_weight = differential_weight
        self.crossover_probability = crossover_probability
        self.strategy = strategy

    def set_parameters(self, differential_weight=0.9, crossover_probability=0.85, strategy=cross_best1, **kwargs):
        r"""Set core parameters of HybridBatAlgorithm algorithm.

        Args:
            differential_weight (Optional[float]): Scaling factor for local search.
            crossover_probability (Optional[float]): Probability of crossover for local search.
            strategy (Optional[Callable[[numpy.ndarray, int, numpy.ndarray, float, float, numpy.random.Generator, Dict[str, Any], numpy.ndarray]]): Local search method based of Differential evolution strategy.

        See Also:
            * :func:`niapy.algorithms.basic.BatAlgorithm.set_parameters`

        """
        super().set_parameters(**kwargs)
        self.differential_weight = differential_weight
        self.crossover_probability = crossover_probability
        self.strategy = strategy

    def get_parameters(self):
        r"""Get parameters of the algorithm.

        Returns:
            Dict[str, Any]: Parameters of the algorithm.

        See Also:
            * :func:`niapy.algorithms.modified.AdaptiveBatAlgorithm.get_parameters`

        """
        d = super().get_parameters()
        d.update({
            'differential_weight': self.differential_weight,
            'crossover_probability': self.crossover_probability
        })
        return d

    def local_search(self, best, loudness, task, i=None, population=None, **kwargs):
        r"""Improve the best solution.

        Args:
            best (numpy.ndarray): Global best individual.
            loudness (float): Loudness.
            task (Task): Optimization task.
            i (int): Index of current individual.
            population (numpy.ndarray): Current best population.

        Returns:
            numpy.ndarray: New solution based on global best individual.

        """
        return task.repair(self.strategy(population, i, self.differential_weight, self.crossover_probability, rng=self.rng, x_b=best), rng=self.rng)



# encoding=utf8

"""Implementations of Schaffer functions."""

import numpy as np
from niapy.problems.problem import Problem

__all__ = ['SchafferN2', 'SchafferN4', 'ExpandedSchaffer']


class SchafferN2(Problem):
    r"""Implementations of Schaffer N. 2 functions.

    Date: 2018

    Author: Klemen Berkovič

    License: MIT

    Function:
    **Schaffer N. 2 Function**
    :math:`f(\textbf{x}) = 0.5 + \frac{ \sin^2 \left( x_1^2 - x_2^2 \right) - 0.5 }{ \left( 1 + 0.001 \left(  x_1^2 + x_2^2 \right) \right)^2 }`

    **Input domain:**
    The function can be defined on any input domain but it is usually
    evaluated on the hypercube :math:`x_i ∈ [-100, 100]`, for all :math:`i = 1, 2,..., D`.

    **Global minimum:** :math:`f(x^*) = 0`, at :math:`x^* = (420.968746,...,420.968746)`

    LaTeX formats:
        Inline:
            $f(\textbf{x}) = 0.5 + \frac{ \sin^2 \left( x_1^2 - x_2^2 \right) - 0.5 }{ \left( 1 + 0.001 \left(  x_1^2 + x_2^2 \right) \right)^2 }$

        Equation:
            \begin{equation} f(\textbf{x}) = 0.5 + \frac{ \sin^2 \left( x_1^2 - x_2^2 \right) - 0.5 }{ \left( 1 + 0.001 \left(  x_1^2 + x_2^2 \right) \right)^2 } \end{equation}

        Domain:
            $-100 \leq x_i \leq 100$

    Reference:
        http://www5.zzu.edu.cn/__local/A/69/BC/D3B5DFE94CD2574B38AD7CD1D12_C802DAFE_BC0C0.pdf

    """

    def __init__(self, lower=-100.0, upper=100.0, *args, **kwargs):
        r"""Initialize SchafferN2 problem..

        Args:
            lower (Optional[Union[float, Iterable[float]]]): Lower bounds of the problem.
            upper (Optional[Union[float, Iterable[float]]]): Upper bounds of the problem.

        See Also:
            :func:`niapy.problems.Problem.__init__`

        """
        kwargs.pop('dimension', None)
        super().__init__(2, lower, upper, *args, **kwargs)

    @staticmethod
    def latex_code():
        r"""Return the latex code of the problem.

        Returns:
            str: Latex code.

        """
        return r'''$f(\textbf{x}) = 0.5 + \frac{ \sin^2 \left( x_1^2 - x_2^2 \right) - 0.5 }{ \left( 1 + 0.001 \left(  x_1^2 + x_2^2 \right) \right)^2 }$'''

    def _evaluate(self, x):
        return 0.5 + (np.sin(x[0] ** 2 - x[1] ** 2) ** 2 - 0.5) / (1 + 0.001 * (x[0] ** 2 + x[1] ** 2)) ** 2


class SchafferN4(Problem):
    r"""Implementations of Schaffer N. 2 functions.

    Date: 2018

    Author: Klemen Berkovič

    License: MIT

    Function:
    **Schaffer N. 2 Function**
    :math:`f(\textbf{x}) = 0.5 + \frac{ \cos^2 \left( \sin \left( x_1^2 - x_2^2 \right) \right)- 0.5 }{ \left( 1 + 0.001 \left(  x_1^2 + x_2^2 \right) \right)^2 }`

    **Input domain:**
        The function can be defined on any input domain but it is usually
        evaluated on the hypercube :math:`x_i ∈ [-100, 100]`, for all :math:`i = 1, 2,..., D`.

    **Global minimum:** :math:`f(x^*) = 0`, at :math:`x^* = (420.968746,...,420.968746)`

    LaTeX formats:
        Inline:
            $f(\textbf{x}) = 0.5 + \frac{ \cos^2 \left( \sin \left( x_1^2 - x_2^2 \right) \right)- 0.5 }{ \left( 1 + 0.001 \left(  x_1^2 + x_2^2 \right) \right)^2 }$

        Equation:
            \begin{equation} f(\textbf{x}) = 0.5 + \frac{ \cos^2 \left( \sin \left( x_1^2 - x_2^2 \right) \right)- 0.5 }{ \left( 1 + 0.001 \left(  x_1^2 + x_2^2 \right) \right)^2 } \end{equation}

        Domain:
            $-100 \leq x_i \leq 100$

    Reference:
        http://www5.zzu.edu.cn/__local/A/69/BC/D3B5DFE94CD2574B38AD7CD1D12_C802DAFE_BC0C0.pdf

    """

    def __init__(self, lower=-100.0, upper=100.0, *args, **kwargs):
        r"""Initialize SchafferN4 problem..

        Args:
            lower (Optional[Union[float, Iterable[float]]]): Lower bounds of the problem.
            upper (Optional[Union[float, Iterable[float]]]): Upper bounds of the problem.

        See Also:
            :func:`niapy.problems.Problem.__init__`

        """
        kwargs.pop('dimension', None)
        super().__init__(2, lower, upper, *args, **kwargs)

    @staticmethod
    def latex_code():
        r"""Return the latex code of the problem.

        Returns:
            str: Latex code.

        """
        return r'''$f(\textbf{x}) = 0.5 + \frac{ \cos^2 \left( \sin \left( x_1^2 - x_2^2 \right) \right)- 0.5 }{ \left( 1 + 0.001 \left(  x_1^2 + x_2^2 \right) \right)^2 }$'''

    def _evaluate(self, x):
        return 0.5 + (np.cos(np.sin(x[0] ** 2 - x[1] ** 2)) ** 2 - 0.5) / (1 + 0.001 * (x[0] ** 2 + x[1] ** 2)) ** 2


class ExpandedSchaffer(Problem):
    r"""Implementations of Expanded Schaffer functions.

    Date: 2018

    Author: Klemen Berkovič

    License: MIT

    Function:
        **Expanded Schaffer Function**
        :math:`f(\textbf{x}) = g(x_D, x_1) + \sum_{i=2}^D g(x_{i - 1}, x_i) \\ g(x, y) = 0.5 + \frac{\sin \left(\sqrt{x^2 + y^2} \right)^2 - 0.5}{\left( 1 + 0.001 (x^2 + y^2) \right)}^2`

    **Input domain:**
        The function can be defined on any input domain but it is usually
        evaluated on the hypercube :math:`x_i ∈ [-100, 100]`, for all :math:`i = 1, 2,..., D`.

    **Global minimum:** :math:`f(x^*) = 0`, at :math:`x^* = (420.968746,...,420.968746)`

    LaTeX formats:
        Inline:
            $f(\textbf{x}) = g(x_D, x_1) + \sum_{i=2}^D g(x_{i - 1}, x_i) \\ g(x, y) = 0.5 + \frac{\sin \left(\sqrt{x^2 + y^2} \right)^2 - 0.5}{\left( 1 + 0.001 (x^2 + y^2) \right)}^2$

        Equation:
            \begin{equation} f(\textbf{x}) = g(x_D, x_1) + \sum_{i=2}^D g(x_{i - 1}, x_i) \\ g(x, y) = 0.5 + \frac{\sin \left(\sqrt{x^2 + y^2} \right)^2 - 0.5}{\left( 1 + 0.001 (x^2 + y^2) \right)}^2 \end{equation}

        Domain:
            $-100 \leq x_i \leq 100$

    Reference:
        http://www5.zzu.edu.cn/__local/A/69/BC/D3B5DFE94CD2574B38AD7CD1D12_C802DAFE_BC0C0.pdf

    """

    def __init__(self, dimension=4, lower=-100.0, upper=100.0, *args, **kwargs):
        r"""Initialize Expanded Schaffer problem..

        Args:
            dimension (Optional[int]): Dimension of the problem.
            lower (Optional[Union[float, Iterable[float]]]): Lower bounds of the problem.
            upper (Optional[Union[float, Iterable[float]]]): Upper bounds of the problem.

        See Also:
            :func:`niapy.problems.Problem.__init__`

        """
        super().__init__(dimension, lower, upper, *args, **kwargs)

    @staticmethod
    def latex_code():
        r"""Return the latex code of the problem.

        Returns:
            str: Latex code.

        """
        return r'''$f(\textbf{x}) = g(x_D, x_1) + \sum_{i=2}^D g(x_{i - 1}, x_i) \\ g(x, y) = 0.5 + \frac{\sin \left(\sqrt{x^2 + y^2} \right)^2 - 0.5}{\left( 1 + 0.001 (x^2 + y^2) \right)}^2$'''

    def _evaluate(self, x):
        x_next = np.roll(x, -1)
        tmp = x ** 2 + x_next ** 2
        val = 0.5 + (np.sin(np.sqrt(tmp)) ** 2 - 0.5) / (1 + 0.001 * tmp) ** 2
        return np.sum(val)



# encoding=utf8
import logging

import numpy as np

from niapy.algorithms.algorithm import Algorithm

logging.basicConfig()
logger = logging.getLogger('niapy.algorithms.basic.SineCosineAlgorithm')
logger.setLevel('INFO')

__all__ = ['SineCosineAlgorithm']


class SineCosineAlgorithm(Algorithm):
    r"""Implementation of sine cosine algorithm.

    Algorithm:
        Sine Cosine Algorithm

    Date:
        2018

    Authors:
        Klemen Berkovič

    License:
        MIT

    Reference URL:
        https://www.sciencedirect.com/science/article/pii/S0950705115005043

    Reference paper:
        Seyedali Mirjalili, SCA: A Sine Cosine Algorithm for solving optimization problems, Knowledge-Based Systems, Volume 96, 2016, Pages 120-133, ISSN 0950-7051, https://doi.org/10.1016/j.knosys.2015.12.022.

    Attributes:
        Name (List[str]): List of string representing algorithm names.
        a (float): Parameter for control in :math:`r_1` value
        r_min (float): Minimum value for :math:`r_3` value
        r_max (float): Maximum value for :math:`r_3` value

    See Also:
        * :class:`niapy.algorithms.Algorithm`

    """

    Name = ['SineCosineAlgorithm', 'SCA']

    @staticmethod
    def info():
        r"""Get basic information of algorithm.

        Returns:
            str: Basic information of algorithm.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""Seyedali Mirjalili, SCA: A Sine Cosine Algorithm for solving optimization problems, Knowledge-Based Systems, Volume 96, 2016, Pages 120-133, ISSN 0950-7051, https://doi.org/10.1016/j.knosys.2015.12.022."""

    def __init__(self, population_size=25, a=3, r_min=0, r_max=2, *args, **kwargs):
        """Initialize SineCosineAlgorithm.

        Args:
            population_size (Optional[int]): Number of individual in population
            a (Optional[float]): Parameter for control in :math:`r_1` value
            r_min (Optional[float]): Minimum value for :math:`r_3` value
            r_max (Optional[float]): Maximum value for :math:`r_3` value

        See Also:
            * :func:`niapy.algorithms.algorithm.Algorithm.__init__`

        """
        super().__init__(population_size, *args, **kwargs)
        self.a = a
        self.r_min = r_min
        self.r_max = r_max

    def set_parameters(self, population_size=25, a=3, r_min=0, r_max=2, **kwargs):
        r"""Set the arguments of an algorithm.

        Args:
            population_size (Optional[int]): Number of individual in population
            a (Optional[float]): Parameter for control in :math:`r_1` value
            r_min (Optional[float]): Minimum value for :math:`r_3` value
            r_max (Optional[float]): Maximum value for :math:`r_3` value

        See Also:
            * :func:`niapy.algorithms.algorithm.Algorithm.set_parameters`

        """
        super().set_parameters(population_size=population_size, **kwargs)
        self.a = a
        self.r_min = r_min
        self.r_max = r_max

    def get_parameters(self):
        r"""Get algorithm parameters values.

        Returns:
            Dict[str, Any]:

        See Also:
            * :func:`niapy.algorithms.algorithm.Algorithm.get_parameters`

        """
        d = super().get_parameters()
        d.update({
            'a': self.a,
            'r_min': self.r_min,
            'r_max': self.r_max
        })
        return d

    def next_position(self, x, best_x, r1, r2, r3, r4, task):
        r"""Move individual to new position in search space.

        Args:
            x (numpy.ndarray): Individual represented with components.
            best_x (numpy.ndarray): Best individual represented with components.
            r1 (float): Number dependent on algorithm iteration/generations.
            r2 (float): Random number in range of 0 and 2 * PI.
            r3 (float): Random number in range [r_min, r_max].
            r4 (float): Random number in range [0, 1].
            task (Task): Optimization task.

        Returns:
            numpy.ndarray: New individual that is moved based on individual ``x``.

        """
        return task.repair(x + r1 * (np.sin(r2) if r4 < 0.5 else np.cos(r2)) * np.fabs(r3 * best_x - x), rng=self.rng)

    def run_iteration(self, task, population, population_fitness, best_x, best_fitness, **params):
        r"""Core function of Sine Cosine Algorithm.

        Args:
            task (Task): Optimization task.
            population (numpy.ndarray): Current population individuals.
            population_fitness (numpy.ndarray[float]): Current population individuals function/fitness values.
            best_x (numpy.ndarray): Current best solution to optimization task.
            best_fitness (float): Current best function/fitness value.
            params (Dict[str, Any]): Additional parameters.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, float, Dict[str, Any]]:
                1. New population.
                2. New populations fitness/function values.
                3. New global best solution.
                4. New global best fitness/objective value.
                5. Additional arguments.

        """
        r1 = self.a - (task.iters + 1) * (self.a / (task.iters + 1))
        r2 = self.uniform(0, 2 * np.pi)
        r3 = self.uniform(self.r_min, self.r_max)
        r4 = self.random()
        population = np.apply_along_axis(self.next_position, 1, population, best_x, r1, r2, r3, r4, task)
        population_fitness = np.apply_along_axis(task.eval, 1, population)
        best_x, best_fitness = self.get_best(population, population_fitness, best_x, best_fitness)
        return population, population_fitness, best_x, best_fitness, {}



# encoding=utf8

"""Implementation of Griewank funcion."""

import numpy as np
from niapy.problems.problem import Problem

__all__ = ['Griewank', 'ExpandedGriewankPlusRosenbrock']


class Griewank(Problem):
    r"""Implementation of Griewank function.

    Date: 2018

    Authors: Iztok Fister Jr. and Lucija Brezočnik

    License: MIT

    Function: **Griewank function**

        :math:`f(\mathbf{x}) = \sum_{i=1}^D \frac{x_i^2}{4000} - \prod_{i=1}^D \cos(\frac{x_i}{\sqrt{i}}) + 1`

        **Input domain:**
        The function can be defined on any input domain but it is usually
        evaluated on the hypercube :math:`x_i ∈ [-100, 100]`, for all :math:`i = 1, 2,..., D`.

        **Global minimum:** :math:`f(x^*) = 0`, at :math:`x^* = (0,...,0)`

    LaTeX formats:
        Inline:
            $f(\mathbf{x}) = \sum_{i=1}^D \frac{x_i^2}{4000} -
            \prod_{i=1}^D \cos(\frac{x_i}{\sqrt{i}}) + 1$

        Equation:
            \begin{equation} f(\mathbf{x}) = \sum_{i=1}^D \frac{x_i^2}{4000} -
            \prod_{i=1}^D \cos(\frac{x_i}{\sqrt{i}}) + 1 \end{equation}

        Domain:
            $-100 \leq x_i \leq 100$

    Reference paper:
    Jamil, M., and Yang, X. S. (2013).
    A literature survey of benchmark functions for global optimisation problems.
    International Journal of Mathematical Modelling and Numerical Optimisation,
    4(2), 150-194.

    """

    def __init__(self, dimension=4, lower=-100.0, upper=100.0, *args, **kwargs):
        r"""Initialize Griewank problem..

        Args:
            dimension (Optional[int]): Dimension of the problem.
            lower (Optional[Union[float, Iterable[float]]]): Lower bound of the problem.
            upper (Optional[Union[float, Iterable[float]]]): Upper bound of the problem.

        See Also:
            :func:`niapy.problems.Problem.__init__`

        """
        super().__init__(dimension, lower, upper, *args, **kwargs)

    @staticmethod
    def latex_code():
        r"""Return the latex code of the problem.

        Returns:
            str: Latex code.

        """
        return r'''$f(\mathbf{x}) = \sum_{i=1}^D \frac{x_i^2}{4000} - \prod_{i=1}^D \cos(\frac{x_i}{\sqrt{i}}) + 1$'''

    def _evaluate(self, x):
        val1 = np.sum(x * x / 4000.0)
        i = np.arange(1, self.dimension + 1)
        val2 = np.product(np.cos(x / np.sqrt(i)))
        return val1 - val2 + 1.0


class ExpandedGriewankPlusRosenbrock(Problem):
    r"""Implementation of Expanded Griewank's plus Rosenbrock function.

    Date: 2018

    Author: Klemen Berkovič

    License: MIT

    Function: **Expanded Griewank's plus Rosenbrock function**

        :math:`f(\textbf{x}) = h(g(x_D, x_1)) + \sum_{i=2}^D h(g(x_{i - 1}, x_i)) \\ g(x, y) = 100 (x^2 - y)^2 + (x - 1)^2 \\ h(z) = \frac{z^2}{4000} - \cos \left( \frac{z}{\sqrt{1}} \right) + 1`

        **Input domain:**
        The function can be defined on any input domain but it is usually
        evaluated on the hypercube :math:`x_i ∈ [-100, 100]`, for all :math:`i = 1, 2,..., D`.

        **Global minimum:**
        :math:`f(x^*) = 0`, at :math:`x^* = (420.968746,...,420.968746)`

    LaTeX formats:
        Inline:
            $f(\textbf{x}) = h(g(x_D, x_1)) + \sum_{i=2}^D h(g(x_{i - 1}, x_i)) \\ g(x, y) = 100 (x^2 - y)^2 + (x - 1)^2 \\ h(z) = \frac{z^2}{4000} - \cos \left( \frac{z}{\sqrt{1}} \right) + 1$

        Equation:
            \begin{equation} f(\textbf{x}) = h(g(x_D, x_1)) + \sum_{i=2}^D h(g(x_{i - 1}, x_i)) \\ g(x, y) = 100 (x^2 - y)^2 + (x - 1)^2 \\ h(z) = \frac{z^2}{4000} - \cos \left( \frac{z}{\sqrt{1}} \right) + 1 \end{equation}

        Domain:
            $-100 \leq x_i \leq 100$

    Reference:
        http://www5.zzu.edu.cn/__local/A/69/BC/D3B5DFE94CD2574B38AD7CD1D12_C802DAFE_BC0C0.pdf

    """

    def __init__(self, dimension=4, lower=-100.0, upper=100.0, *args, **kwargs):
        r"""Initialize Expanded Griewank's plus Rosenbrock problem..

        Args:
            dimension (Optional[int]): Dimension of the problem.
            lower (Optional[Union[float, Iterable[float]]]): Lower bounds of the problem.
            upper (Optional[Union[float, Iterable[float]]]): Upper bounds of the problem.

        See Also:
            :func:`niapy.problems.Problem.__init__`

        """
        super().__init__(dimension, lower, upper, *args, **kwargs)

    @staticmethod
    def latex_code():
        r"""Return the latex code of the problem.

        Returns:
            str: Latex code.

        """
        return r'''$f(\textbf{x}) = h(g(x_D, x_1)) + \sum_{i=2}^D h(g(x_{i - 1}, x_i)) \\ g(x, y) = 100 (x^2 - y)^2 + (x - 1)^2 \\ h(z) = \frac{z^2}{4000} - \cos \left( \frac{z}{\sqrt{1}} \right) + 1$'''

    def _evaluate(self, x):
        x1 = 100.0 * (x[1:] - x[:-1] ** 2.0) ** 2.0 + (1 - x[:-1]) ** 2.0
        x2 = x1 * x1 / 4000.0 - np.cos(x1 / np.sqrt(np.arange(1, self.dimension)))
        return np.sum(x2)



# encoding=utf8
import logging

import numpy as np

from niapy.algorithms.algorithm import Algorithm

logging.basicConfig()
logger = logging.getLogger('niapy.algorithms.other')
logger.setLevel('INFO')

__all__ = ['RandomSearch']


class RandomSearch(Algorithm):
    r"""Implementation of a simple Random Algorithm.

    Algorithm:
        Random Search

    Date:
        11.10.2020

    Authors:
        Iztok Fister Jr., Grega Vrbančič

    License:
        MIT

    Reference URL: https://en.wikipedia.org/wiki/Random_search

    Attributes:
        Name (List[str]): List of strings representing algorithm name.

    See Also:
        * :class:`niapy.algorithms.Algorithm`

    """

    Name = ['RandomSearch', 'RS']

    @staticmethod
    def info():
        r"""Get basic information of algorithm.

        Returns:
            str: Basic information of algorithm.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""None"""

    def __init__(self, *args, **kwargs):
        """Initialize RandomSearch."""
        kwargs.pop('population_size', None)
        super().__init__(1, *args, **kwargs)
        self.candidates = None

    def set_parameters(self, **kwargs):
        r"""Set the algorithm parameters/arguments.

        See Also
            * :func:`niapy.algorithms.Algorithm.set_parameters`

        """
        kwargs.pop('population_size', None)
        Algorithm.set_parameters(self, population_size=1, **kwargs)
        self.candidates = None

    def get_parameters(self):
        r"""Get algorithms parameters values.

        Returns:
            Dict[str, Any]:
        See Also
            * :func:`niapy.algorithms.Algorithm.get_parameters`

        """
        d = Algorithm.get_parameters(self)
        return d

    def init_population(self, task):
        r"""Initialize the starting population.

        Args:
            task (Task): Optimization task.
        Returns:
            Tuple[numpy.ndarray, float, dict]:
            1. Initial solution
            2. Initial solutions fitness/objective value
            3. Additional arguments

        """
        if task.max_iters != np.inf:
            total_candidates = task.max_iters
        elif task.max_evals != np.inf:
            total_candidates = task.max_evals
        else:
            total_candidates = 0
        self.candidates = []
        x = None
        for i in range(total_candidates):
            while True:
                x = task.lower + task.range * self.random(task.dimension)
                if not np.any([np.all(a == x) for a in self.candidates]):
                    self.candidates.append(x)
                    break

        x_fit = task.eval(self.candidates[0])
        return x, x_fit, {}

    def run_iteration(self, task, x, x_fit, best_x, best_fitness, **params):
        r"""Core function of the algorithm.

        Args:
            task (Task):
            x (numpy.ndarray):
            x_fit (float):
            best_x (numpy.ndarray):
            best_fitness (float):
            **params (dict): Additional arguments.

        Returns:
            Tuple[numpy.ndarray, float, numpy.ndarray, float, dict]:
            1. New solution
            2. New solutions fitness/objective value
            3. New global best solution
            4. New global best solutions fitness/objective value
            5. Additional arguments

        """
        current_candidate = task.iters if task.max_iters != np.inf else task.evals
        x = self.candidates[current_candidate]
        x_fit = task.eval(x)
        best_x, best_fitness = self.get_best(x, x_fit, best_x, best_fitness)
        return x, x_fit, best_x, best_fitness, {}
